"""
A module for instantiating clusters for use in ClusterGraph objects.
"""
# Standard imports
import uuid
import copy

# Local imports
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL
FIX_NON_PSD_MATRICES = False

# TODO: add evidence observation functionality
#   (perhaps only important for non-linear gaussian and similar approximate transformation factors)

# pylint: disable=protected-access


class Cluster:
    """
    A class for instantiating clusters for use in ClusterGraph objects.
    """

    def __init__(self, factor, cluster_name_prefix=""):
        """
        Construct a cluster with an initial factor potential.

        :param Factor factor: The factor to use in the cluster
        :param str cluster_name_prefix: The name of the cluster.
        """
        self._factor = factor
        # self._cluster_id = name if name is not None else str(uuid.uuid1())
        # self._cluster_id = name if name is not None else str(factor.var_names)
        self._cluster_id = cluster_name_prefix + ",".join(factor.var_names)
        self._neighbour_sepsets = {}  # neighbour_id as key
        self._received_message_factors = {}  # neighbour_id as key
        self._neighbour_references = []
        self._outward_message_paths = []
        self._special_evidence = None

    def add_special_evidence(self, special_evidence):
        """
        Add special evidence to the cluster. This evidence will not be used to reduce the cluster potential, but will be
        used in the calculation of messages from the cluster. This is used for clusters with factors of which the joint
        distribution is approximated based on conditioning distributions (i.e the NonLinearGaussian factor).

        :param dict special_evidence: The special evidence (i.e {'a':0.3, 'b':2.0})
        """
        self._special_evidence = special_evidence

    def remove_all_neighbours(self):
        """
        Remove all neighbours.
        """
        self._neighbour_sepsets = {}  # neighbour_id as key

    def add_neighbour(self, other_cluster, sepset):
        """
        Add a neighbour to this cluster's list of neighbours.

        :param Cluster other_cluster: The cluster to add as a neighbour
        :param list sepset: The sepset between the neighbours (can be smaller than the scope intersection in order
                       to enforce the running intersection property)
        """
        # TODO: Add var_names check here (there should be a non-zero intersection)
        assert len(sepset) > 0, "Error: cant add neighbour with no overlapping variable scope."
        self._neighbour_references.append(other_cluster)
        self._neighbour_sepsets[other_cluster.cluster_id] = list(sepset)

    # TODO: merge this with add_neighbour
    def add_outward_message_path(self, outward_message_path):
        """
        Add an outward graph message path.

        :param outward_message_path: The graph message path to add.
        """
        self._outward_message_paths.append(outward_message_path)

    def make_message(self, neighbour_id):
        """
        Make a Message to send to the neighbour with a specified id.

        :param neighbour_id: The specified id corresponding to the neighbour cluster.
        :return: The Message
        :rtype: Message
        """
        sepset_vars = self._neighbour_sepsets[neighbour_id]
        message_factor = self._factor.copy()
        if self._special_evidence is not None:
            evidence_vrs = list(self._special_evidence.keys())
            evidence_values = list(self._special_evidence.values())
            message_factor = message_factor.reduce(evidence_vrs, evidence_values)
        message_factor = message_factor.marginalize(sepset_vars, keep=True)
        if neighbour_id in self._received_message_factors:
            prev_received_message_factor = self._received_message_factors[neighbour_id]
            message_factor = message_factor.cancel(prev_received_message_factor)
        message = Message(factor=message_factor, sender_id=self.cluster_id, receiver_id=neighbour_id)
        return message

    def receive_message(self, message):
        """
        Receive Message.

        :param Message message: The received Message
        """
        # Absorb message
        assert message.receiver_id == self.cluster_id, "Error: Message not meant for this Cluster."
        self._factor = self._factor.multiply(message.factor)

        # Cancel out any message previously received from sender cluster
        sender_id = message.sender_id
        if sender_id in self._received_message_factors:
            prev_received_message_factor = self._received_message_factors[sender_id]
            self._factor = self._factor.cancel(prev_received_message_factor)

        self._received_message_factors[message.sender_id] = message.factor.copy()

    def get_sepset(self, neighbour_id):
        """
        Get the sepset between this cluster and another.

        :param str neighbour_id: The id of the other cluster.
        :return: The sepset
        :rtype: list
        """
        sepset = self._neighbour_sepsets[neighbour_id].copy()
        return sepset

    @property
    def neighbour_ids(self):
        """
        Get a list of ids corresponding to the cluster's neigbours.
        """
        return list(self._neighbour_sepsets.keys())

    @property
    def var_names(self):
        """
        The variable names associated with the cluster factor.

        :return: the var_names
        """
        return self._factor.var_names

    @property
    def cluster_id(self):
        """
        The cluster's id.

        :return: The cluster_id
        """
        return self._cluster_id

    @property
    def factor(self):
        """
        The factor.

        :return: The factor.
        """
        return self._factor.copy()


class Message:

    """
    A class for instantiating Message objects.
    """

    def __init__(self, factor, sender_id, receiver_id):
        """
        The initializer.

        :param factor: The Message factor.
        :param sender_id:  The cluster_id of the sender cluster
        :param receiver_id: The cluster_id of the receiver cluster
        """
        self.uuid = str(uuid.uuid1())
        self._sender_id = sender_id
        self._receiver_id = receiver_id
        self._factor = factor

    def equals(self, other, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this message equals another message.

        :param other: The other message.
        :param float rtol: The relative tolerance to use for factor equality check.
        :param float atol: The absolute tolerance to use for factor equality check.
        :return: The result of the equality check.
        :rtype: bool
        """
        if self._sender_id != other.sender_id:
            return False
        if self._receiver_id != other._receiver_id:
            return False
        if not self._factor.equals(other._factor, rtol=rtol, atol=atol):
            return False
        return True

    def __repr__(self):
        """
        Get the string representation of the message.

        :return: The string.
        :rtype: str
        """
        repr_str = (
            "sender_id: "
            + self._sender_id
            + "\n"
            + "receiver_id: "
            + self._receiver_id
            + "\n"
            + "factor: \n"
            + self._factor.__repr__()
        )
        return repr_str

    @property
    def sender_id(self):
        """
        The Message's sender id

        :return: The sender_id
        """
        return self._sender_id

    @property
    def receiver_id(self):
        """
        The Message's receiver id

        :return: The receiver_id
        """
        return self._receiver_id

    @property
    def factor(self):
        """
        The Message's factor.

        :return: The factor
        """
        return self._factor.copy()

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between the message factor and a vacuous (uniform) version of it.

        :return: The KL-divergence
        """
        return self.factor.distance_from_vacuous()

    def kl_divergence(self, message):
        """
        Get the KL-divergence D_KL(self.factor || message.factor) between a this message factor and another message
        factor.

        :param message: The other message of which the factor will be used to compare to this message's factor.
        :return: The KL-divergence
        """
        distance = self.factor.kl_divergence(message.factor)
        return distance

    @property
    def var_names(self):
        """
        The variable names associated with the cluster factor.

        :return: the var_names
        """
        return self._factor.var_names

    def copy(self):
        """
        Copy this Message.

        :return: The message copy.
        :rtype: Message
        """
        msg_copy = Message(self.factor.copy(), copy.deepcopy(self.sender_id), copy.deepcopy(self.receiver_id))
        return msg_copy



"""
A module for instantiating and performing operations on Non-linear Gaussian functions.
"""

# Standard imports
import operator
import copy

# Third-party imports
import numpy as np

# Local imports
from veroku.factors import _factor_utils
from veroku.factors._sigma_points import get_sigma_points_array, sigma_points_array_to_joint_params
from veroku.factors._factor import Factor
from veroku.factors.gaussian import Gaussian
from veroku.factors.experimental.gaussian_mixture import GaussianMixture
from veroku.factors._factor_utils import (
    make_square_matrix,
    indexed_square_matrix_operation,
    format_list_elements,
)
from veroku.factors._factor_template import FactorTemplate
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL


# TODO: see that lazy evaluation (recompute joint) is done sensibly and consistently

# pylint: disable=protected-access


class NonLinearGaussian(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self,
                 conditioning_vars,
                 conditional_vars,
                 transform,
                 noise_cov,
                 log_weight=0.0,
                 joint_distribution=None,
                 conditional_update_factor=None,
                 conditioning_factor=None,
                 observed_evidence=None):
        """
        The initializer.

        :param list conditioning_vars: The conditioning vars.
        :param list conditional_vars: The conditioning vars.
        :param transform: The transformation that links the conditioning to the conditional distribution
        :param noise_cov: The noise covariance matrix of the zero mean Gaussian noise.
        :param Gaussian conditional_update_factor: A factor, with same scope, subset of, the conditional vars, to be
            multiplied with the transformed (joint) distribution to form the final joint distribution.
        :param Gaussian conditioning_factor: The factor with scope less than or equals to the conditional variables to be
            transformed by the transform. If the scope is less than the conditional variables set, observed evidence
            (if available) can be used to still enable the successful transformation and calculation of the joint.
        :param dict observed_evidence: A dictionary mapping variable names ot observed values.
        """
        if len(set(conditioning_vars).intersection(conditional_vars)) != 0:
            raise ValueError("variable overlap between conditioning_vars and conditional_vars.")
        var_names = conditioning_vars + conditional_vars

        super().__init__(var_names=var_names)
        self.transform = transform
        if noise_cov.shape[0] != len(conditional_vars):
            raise ValueError(
                "noise_cov rows and columns should correspond to the number of conditional variables"
            )
        self.noise_cov = make_square_matrix(noise_cov)

        self.log_weight = log_weight
        self.conditioning_vars = conditioning_vars.copy()
        self.conditional_vars = conditional_vars.copy()

        self.conditioning_factor = None
        if conditioning_factor is not None:
            self.conditioning_factor = conditioning_factor.copy()

        self.conditional_update_factor = None
        if conditional_update_factor is not None:
            self.conditional_update_factor = conditional_update_factor.copy()

        if joint_distribution is not None:
            self._joint_distribution = joint_distribution.copy()
        else:
            self._joint_distribution = Gaussian.make_vacuous(var_names)
        if observed_evidence is None:
            self.observed_evidence = {}
        else:
            self.observed_evidence = copy.deepcopy(observed_evidence)

    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: Gaussian
        :return: The result of the comparison.
        :rtype: bool
        """
        return self._joint_distribution.equals(factor, rtol=rtol, atol=atol)

    def distance_from_vacuous(self):
        """
        NOTE: Not Implemented yet.
        Get the Kullback-Leibler (KL) divergence between the message factor and a vacuous (uniform) version of it.

        :return: The KL-divergence
        """
        raise NotImplementedError('This function has not been implemented yet.')

    def kl_divergence(self, factor):
        """
        NOTE: Not Implemented yet.
        Get the KL-divergence D_KL(self || factor) between a normalized version of this factor and another factor.

        :param factor: The other factor
        :type factor: GaussianMixture
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        raise NotImplementedError('This function has not been implemented yet.')

    def copy(self):
        """
        Make a copy of the factor.

        :return: the copy
        :rtype: NonLinearGaussian
        """
        nlg_copy = NonLinearGaussian(
            conditioning_vars=copy.deepcopy(self.conditioning_vars),
            conditional_vars=copy.deepcopy(self.conditional_vars),
            transform=copy.deepcopy(self.transform),
            noise_cov=copy.deepcopy(self.noise_cov),
            log_weight=self.log_weight,
            joint_distribution=self._joint_distribution,
            conditional_update_factor=self.conditional_update_factor,
            conditioning_factor=self.conditioning_factor,
            observed_evidence=self.observed_evidence,
        )
        return nlg_copy

    @property
    def joint_distribution(self):
        """
        The joint distribution.

        :return: the joint distribution
        """
        self_copy = self.copy()
        self_copy._recompute_joint()
        return self_copy._joint_distribution

    def get_prec(self):
        """
        Get the K matrix (after ensuring that the canonical form is updated.)

        :return: The K parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_prec()

    def get_h(self):
        """
        Get the log h vector (after ensuring that the canonical form is updated.)

        :return: The h parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_h()

    def get_g(self):
        """
        Get the g scalar (after ensuring that the canonical form is updated.)

        :return: The g parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_g()

    def get_cov(self):
        """
        Get the covariance matrix (after ensuring that the covariance form is updated.)

        :return: The cov parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_cov()

    def get_mean(self):
        """
        Get the mean vector (after ensuring that the covariance form is updated.)

        :return: The mean parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_mean()

    def get_log_weight(self):
        """
        Get the log weight (after ensuring that the covariance form is updated.)

        :return: The log_weight parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_log_weight()

    def add_log_weight(self, log_weight_to_add):
        """
        Add a log weight to the factor.

        :param log_weight_to_add: The log weight to add.
        """
        self.log_weight += log_weight_to_add

    def _recompute_joint(self):
        """
        Recompute the joint distribution.
        """
        # TODO: improve this function (after code coverage is completed for this class) and remove disable below.
        # pylint: disable=too-many-locals
        observed_vars = list(self.observed_evidence.keys())

        conditioning_observed_vars = [var for var in observed_vars if var in self.conditioning_vars]
        conditioning_observed_values = [
            self.observed_evidence[var] for var in observed_vars if var in self.conditioning_vars
        ]

        conditional_observed_vars = [var for var in observed_vars if var in self.conditional_vars]
        conditional_observed_values = [
            self.observed_evidence[var] for var in observed_vars if var in self.conditional_vars
        ]

        # under 3 conditions we can successfully perform the transformation:
        # 1. We have a well defined conditional distribution with the full scope of the conditional variables.
        # 2. All the conditional variables are observed.
        # 3. We have a conditional distribution over a subset of the conditional variables and the rest of the
        # conditional variables are observed.

        # Initial feasibility check
        conditioning_factor_var_names = []
        if self.conditioning_factor is not None:
            conditioning_factor_var_names = self.conditioning_factor.var_names
        vars_obs_vars_union = set(conditioning_factor_var_names).union(set(conditioning_observed_vars))
        if vars_obs_vars_union != set(self.conditioning_vars):
            # The conditioning_factor variables and observed conditional variables combined do not make up a
            # complete set - so we cannot calculate the transform.
            self._joint_distribution = Gaussian.make_vacuous(self.var_names)
            return

        # If we are still here, there we can calculate the transformation.
        conditioning_factor_sigma_points_array = np.expand_dims(np.array([]), axis=0).T  # empty dummy for generalisation
        if self.conditioning_factor is not None:
            conditioning_factor = self.conditioning_factor.copy()
            cond_factor_observed_vars = list(set(observed_vars).intersection(conditioning_factor.var_names))
            cond_factor_observed_values = [self.observed_evidence[var] for var in cond_factor_observed_vars]

            if len(cond_factor_observed_vars) > 0:
                conditioning_factor = conditioning_factor.reduce(
                    vrs=cond_factor_observed_vars, values=cond_factor_observed_values
                )

            conditioning_factor_sigma_points_array = get_sigma_points_array(conditioning_factor)

        # extend sigma points with observed variable values
        if len(conditioning_observed_vars) > 0:
            observed_vec = _factor_utils.make_column_vector(conditioning_observed_values)
            tiled_observed_vec = np.tile(observed_vec, [1, conditioning_factor_sigma_points_array.shape[1]])
            extended_sigma_points_array = np.concatenate([tiled_observed_vec, conditioning_factor_sigma_points_array])
            extended_sigma_point_vars = conditioning_observed_vars + conditioning_factor.var_names
        else:
            extended_sigma_points_array = conditioning_factor_sigma_points_array
            extended_sigma_point_vars = conditioning_factor.var_names

        # TODO: Add conditional var_names here somehow. We need to ensure that the correct variables are used
        # in the correct place in the transformation. In the past, we have simply done this by standardising
        # on the variable indices in different place, but this is probably not very safe.
        joint_cov, joint_mean = sigma_points_array_to_joint_params(sigma_points_array=extended_sigma_points_array,
                                                                   transform=self.transform,
                                                                   var_names=extended_sigma_point_vars)
        if (joint_mean.shape[0] != len(self.var_names)) or (joint_cov.shape[0] != len(self.var_names)):
            raise AssertionError("Transform resulted in incorrect number of variables.")

        # remove the observed vars again after transformation
        nov = len(conditioning_observed_vars)  # number of observed vars
        joint_cov = joint_cov[nov:, nov:]
        joint_mean = joint_mean[nov:, :]
        joint_var_names = conditioning_factor.var_names + self.conditional_vars

        joint_cov_plus_noise, joint_vars = indexed_square_matrix_operation(
            joint_cov, self.noise_cov, joint_var_names, self.conditional_vars, operator.add
        )
        joint_log_weight = self.log_weight + conditioning_factor.log_weight
        self._joint_distribution = Gaussian(cov=joint_cov_plus_noise,
                                            mean=joint_mean,
                                            log_weight=joint_log_weight,
                                            var_names=joint_vars,
        )
        if self.conditional_update_factor is not None:
            self._joint_distribution = self._joint_distribution.multiply(self.conditional_update_factor)
        if len(conditional_observed_vars) > 0:
            self._joint_distribution = self._joint_distribution.reduce(
                conditional_observed_vars, conditional_observed_values
            )

    def normalize(self):
        """
        Normalize the factor.

        :return: The normalized factor.
        :rtype: Gaussian
        """
        # TODO: make this more efficient
        return self.joint_distribution.normalize()

    def multiply(self, factor):
        """
        Multiply this factor with another factor.

        :param factor: the factor to multiply with
        :return: the resulting factor
        :rtype: NonLinearGaussian
        """

        if isinstance(factor, GaussianMixture):
            nlgs = []
            for gaussian_factor in factor.components:
                nlgs.append(self.multiply(gaussian_factor))
            return NonLinearGaussianMixture(nlgs)

        self_copy = self.copy()
        if isinstance(factor, NonLinearGaussian):
            raise NotImplementedError(
                "Multiplication of two nonlinear-Gaussian factors has not been implemented."
            )
        if factor._is_vacuous:
            # TODO: check this (esp to see what to do with)
            return self_copy
        # TODO: change to 'update conditional model and re-transform model' (if correct)
        if not isinstance(factor, Gaussian):
            raise ValueError(f"factor must be Gaussian, got {type(factor)}")

        if set(factor.var_names) <= set(self_copy.conditional_vars):
            if self_copy.conditional_update_factor is None:
                self_copy.conditional_update_factor = factor.copy()
            else:
                self_copy.conditional_update_factor = self_copy.conditional_update_factor.multiply(factor)
        elif set(factor.var_names) <= set(self_copy.conditioning_vars):
            if self_copy.conditioning_factor is None:
                self_copy.conditioning_factor = factor.copy()
            else:
                self_copy.conditioning_factor = self_copy.conditioning_factor.multiply(factor)
        else:
            raise ValueError(
                f"cannot absorb factor with scope ({factor.var_names}) \n "
                f"           which has neither conditional ({self_copy.conditioning_vars}) \n "
                f"           nor conditioning ({self_copy.conditional_vars}) scope."
            )
        return self_copy

    def divide(self, factor):
        """
        Divide this factor by another factor.

        :param factor: the factor divide by
        :return: the resulting factor
        :rtype: NonLinearGaussian
        """
        self_copy = self.copy()
        if factor._is_vacuous:
            # TODO: check this (esp to see what to do with)
            return self_copy
        # TODO: change to 'update conditional model and re-transform model' (if correct)
        if not isinstance(factor, Gaussian):
            raise ValueError("factor must be Gaussian.")

        if set(factor.var_names) <= set(self_copy.conditional_vars):
            self_copy.conditional_update_factor = self_copy.conditional_update_factor.divide(factor)
        elif set(factor.var_names) <= set(self_copy.conditioning_vars):
            if self_copy.conditioning_factor is None:
                raise NotImplementedError()
            self_copy.conditioning_factor = self_copy.conditioning_factor.divide(factor)
        else:
            raise ValueError(
                f"cannot cancel factor with scope ({factor.var_names}) \n "
                f"           which has neither conditional ({self_copy.conditioning_vars}) \n "
                f"           nor conditioning ({self_copy.conditional_vars}) scope."
            )
        self_copy._recompute_joint()
        return self_copy

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this factor and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :param values: the values of the observed variables (list or vector-like object)
        :return: the resulting factor
        :rtype: NonLinearGaussian
        """

        self_copy = self.copy()
        evidence_dict = dict(zip(vrs, values))
        self_copy.observed_evidence.update(evidence_dict)
        return self_copy

    def marginalize(self, vrs, keep=True):
        """
        Integrate out variables from this factor.

        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: the resulting marginal
        :rtype: Gaussian
        """

        vrs_to_keep = super().get_marginal_vars(vrs, keep=keep)
        assert set(vrs_to_keep) <= set(self.var_names), "Error: asked to keep vars not in factor."
        self_copy = self.copy()
        self_copy._recompute_joint()
        if self_copy._joint_distribution._is_vacuous:
            if set(vrs_to_keep) <= set(self_copy.conditioning_vars):
                if self_copy.conditioning_factor is not None:
                    if set(vrs_to_keep) <= set(self_copy.conditioning_factor.var_names):
                        marginal = self_copy.conditioning_factor.marginalize(vrs_to_keep, keep=True)
                        return marginal
            if set(vrs_to_keep) <= set(self_copy.conditional_vars):
                if self_copy.conditional_update_factor is not None:
                    if set(vrs_to_keep) <= set(self_copy.conditional_update_factor.var_names):
                        marginal = self_copy.conditional_update_factor.marginalize(vrs_to_keep, keep=True)
                        marginal._add_log_weight(self.log_weight)
                        return marginal
            else:
                return Gaussian.make_vacuous(vrs_to_keep)
        return self_copy._joint_distribution.marginalize(vrs_to_keep, keep=True)

    def sample(self, num_samples):
        """
        Draw samples from the Gaussian joint distribution defined by the non-linear Gaussian and the factors that have
        (potentially) been multiplied with it.

        :param num_samples: The number of samples to draw.
        :type num_samples:
        :return: The samples.
        :rtype: int
        """
        self._recompute_joint()
        return self._joint_distribution.sample(num_samples=num_samples)

    def show(self):
        """
        Print the parameters of the nonlinear Gaussian distribution.
        """
        # TODO: add __repr__ function and use here.
        print("Non-linear Gaussian")
        print("conditioning variables:", self.conditioning_vars)
        print("conditional variables:", self.conditional_vars)
        print("transform: ", self.transform)
        print("noise covariance = \n", self.noise_cov)
        print("joint_distribution = \n")
        if self._joint_distribution is not None:
            self._recompute_joint()
            self._joint_distribution.show()

        print("conditional update distribution = \n")
        if self.conditional_update_factor is not None:
            self.conditional_update_factor.show()
        else:
            print("vacuous")

        print("conditioning distribution = \n")
        if self.conditioning_factor is not None:
            self.conditioning_factor.show()
        else:
            print("vacuous")

    def plot(self):
        """
        Plot the joint distribution.
        """
        self_copy = self.copy()
        self_copy._recompute_joint()
        self._joint_distribution.plot()


class NonLinearGaussianMixture(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    def __init__(self, factors, split_singles_before_absorb=True):
        """
        The initializer.

        :param factors: The list of factors.
        :type factors: NonLinearGaussian factor list
        :param bool split_singles_before_absorb: Whether or not to split a Gaussian (or single component
            GaussianMixture) before it is multiplied in. This can result in a better approximation of non-linear
            Gaussian transformations.
        """
        self.nlgs = []
        if len(factors) == 0:
            raise ValueError("received empty list of factors.")
        super().__init__(var_names=factors[0].var_names)
        for factor in factors:
            if not isinstance(factor, NonLinearGaussian):
                raise TypeError(f"expected NonLinearGaussian type, received {type(factor)}.")
            if not set(self.var_names) == set(factor.var_names):
                raise ValueError(
                    f"Inconsistent variable names. First factor has var_names = {self.var_names},"
                    f" another has var_names = {factor.var_names}"
                )
            self.nlgs.append(factor.copy())
        self.split_singles_before_absorb = split_singles_before_absorb

    def normalize(self):
        """
        Normalise the factor (not yet implemented).
        """
        raise NotImplementedError()

    def copy(self):
        """
        Make a copy of this factor.

        :return: the factor
        :rtype: NonLinearGaussianMixture
        """
        return NonLinearGaussianMixture(self.nlgs)

    @property
    def joint_distribution(self):
        gaussian_joints = []
        for nlg in self.nlgs:
            gaussian_joints.append(nlg.joint_distribution)
        return GaussianMixture(gaussian_joints)

    def multiply(self, factor):
        """
        Multiply this factor with another factor.

        :param factor: the factor to multiply with
        :return: the resulting factor
        :rtype: NonLinearGaussianMixture
        """
        if isinstance(factor, GaussianMixture):
            gm_factor = factor
        elif isinstance(factor, Gaussian):
            gm_factor = GaussianMixture([Gaussian])
        else:
            raise NotImplementedError()

        # TODO: Generalise the Gaussian.split_gaussian function to more than one dimensional cases and remove the
        #  limitation here
        if self.split_singles_before_absorb and len(gm_factor.var_names) == 1:
            if len(gm_factor.components) == 1:
                gm_factor = gm_factor.components[0]._split_gaussian()

        new_nlgs = []
        for gauss in gm_factor.components:
            for nlg in self.nlgs:
                new_nlgs.append(nlg.multiply(gauss))
        return NonLinearGaussianMixture(new_nlgs)

    def divide(self, factor):
        """
        Divide this factor by another factor.

        :param factor: the factor divide by
        :return: the resulting factor
        :rtype: NonLinearGaussianMixture
        """
        if isinstance(factor, Gaussian):
            gaussian_factor = factor
        elif isinstance(factor, GaussianMixture):
            # TODO: Add better Gaussian mixture division approximation
            gaussian_factor = factor.moment_match_to_single_gaussian()
        else:
            raise NotImplementedError("cannot divide NonLinearGaussianMixture by {type(factor)}.")
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.divide(gaussian_factor))
        return NonLinearGaussianMixture(new_nlgs)

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this factor and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :param values: the values of the observed variables (list or vector-like object)
        :return: the resulting factor
        :rtype: NonLinearGaussianMixture
        """
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.reduce(vrs, values))
        return NonLinearGaussianMixture(new_nlgs)

    def marginalize(self, vrs, keep=True):
        """
        Integrate out variables from this factor.

        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: the resulting marginal
        :rtype: NonLinearGaussianMixture
        """
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.marginalize(vrs, keep))
        return NonLinearGaussianMixture(new_nlgs)

    def sample(self, num_samples):
        """
        Draw samples from the Gaussian mixture joint distribution defined by the non-linear Gaussians and the factors
        that have (potentially) been multiplied with them.

        :param num_samples: The number of samples to draw.
        :type num_samples:
        :return: The samples.
        :rtype: int
        """

        # TODO: fix incorrect num samples issue
        sample_sets = []
        for nlg in self.nlgs:
            samples = nlg.sample(num_samples)
            sample_sets.append(samples)
        return np.concatenate(sample_sets)

    def plot(self):
        """
        Plot the joint Gaussian Mixture distribution.
        """
        gaussian_components = [nlg.joint_distribution for nlg in self.nlgs]
        gaussian_mixture = GaussianMixture(gaussian_components)
        gaussian_mixture.plot()

    def show(self):
        """
        Print the parameters of the nonlinear Gaussian distribution.
        """
        for i, nlg in enumerate(self.nlgs):
            print(f"######### NLG {i} #############################")
            nlg.show()


class NonLinearGaussianTemplate(FactorTemplate):
    """
    A template class for NonLinearGaussian factors.
    """

    def __init__(self, conditioning_var_templates, conditional_var_templates, transition_function, noise_cov):
        """
        The initializer.

        :param conditioning_var_templates: The list of formattable strings for the conditioning variables (i.e: ['var_a{i}_{t}', 'var_b{i}_{t}'])
        :param conditional_var_templates: The list of formattable strings for the conditional variables (i.e: ['var_c{i}_{t}', 'var_d{i}_{t}'])
        :param conditioning_var_templates: The formattable strings for the conditioning variables.
        :type conditioning_var_templates: str list
        :param conditional_var_templates: The formattable strings for the conditioning variables.
        :type conditional_var_templates: str list
        :param callable transition_function: The function that specifies the non-linear transform. This function takes
            2 parameters, a vector-like value to be transformed and the variable names list specifying the names of the
            variable elements of the value vector (i.e transition_function = lamda x, var_names: np.square(x)).
            The variable names do not need to be used, but can be useful in certain cases where functions need to be
            applied to specific variables.
        :param noise_cov: The noise covariance matrix of the additive noise random variable.
        """

        super().__init__(
            conditioning_var_templates=conditioning_var_templates,
            conditional_var_templates=conditional_var_templates,
        )
        self.transition_function = transition_function
        self.noise_cov = noise_cov

    def make_factor(self, format_dict=None, conditioning_vars=None, conditional_vars=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :param conditioning_vars: The conditioning variables strings.
        :type conditioning_vars: str list
        :param conditional_vars: The conditioning variables strings.
        :type conditional_vars: str list
        :return: The instantiated factor.
        :rtype: NonLinearGaussianMixture
        """
        if format_dict is not None:
            assert conditioning_vars is None
            assert conditioning_vars is None

            conditioning_vars = format_list_elements(self._conditioning_var_templates, format_dict)
            conditional_vars = format_list_elements(self._conditional_var_templates, format_dict)

        return NonLinearGaussian(
            conditioning_vars,
            conditional_vars,
            self.transition_function,
            self.noise_cov,
            log_weight=0.0,
            joint_distribution=None,
        )



"""
A module for instantiating sparse tables with log probabilities.
"""

# Standard imports
import copy
import itertools
import operator

# Third-party imports
import numpy as np
from scipy import special
import pandas as pd

# Local imports
from veroku.factors._factor import Factor
from veroku.factors._factor_template import FactorTemplate
from veroku.factors import _factor_utils
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL

# pylint: disable=protected-access


def _make_dense(factor):
    """
    Convert a factor to a dense factor by instantiating all of the implicit default value assignments.

    :param factor: The factor to convert.
    :type factor: SparseCategorical
    :return: The dense factor version.
    :rtype: SparseCategorical
    """
    factor_copy = factor.copy()
    dense_assingments = itertools.product(*[range(factor_copy.var_cards[v]) for v in factor_copy.var_names])
    for assign in dense_assingments:
        if assign not in factor_copy.log_probs_table:
            factor_copy.log_probs_table[assign] = factor_copy.default_log_prob
    return factor_copy


def _make_dense_default_probs_dict(cardinalities, default_value):
    """
    Make a dictionary representing a categorical probability table that contains all assignments, as specified by the
    given cardinalities (starting from 0).

    :param cardinalities: The given cardinalities.
    :type cardinalities: list
    :param default_value: The default value to assign the missing values.
    :type default_value: float
    :returns: The dense representation.
    :rtype: dict
    Example:

    .. code-block:: python

        >>> _make_dense_default_probs_dict(cardinalities=[2,3], default_value=0.0)
        {(0,0):0.0,
        (0,1):0.0,
        (0,2):0.0,
        (1,0):0.0,
        (1,1):0.0,
        (1,2):0.0}
    """
    dense_assingments = itertools.product(*[range(c) for c in cardinalities])
    dense_default_dict = dict(zip(dense_assingments, [default_value] * np.product(cardinalities)))
    return dense_default_dict


def _make_inner_dense(sparse_nested_table, outer_inner_cards, default_value):
    """
    Convert n sparse nested table dictionary's implicit default values in the innetables to real entries so that
    all possible assignments are present in the inner sub tables.

    :param sparse_nested_table: The nested dictionary representing a sparse table.
    :type sparse_nested_table: dictionary of dictionaries
    :param outer_inner_cards: The lists of the cardinalities for the outer and inner part of the nested table respectively
        i.e [[2,2], [3, 4]] if the table dict is {(0,0):{(2,3):0.3},....} where the outer variables has cardinalities [2,2]
        and the inner ones [3,4]
    """
    sparse_nested_inner_dense_table = copy.deepcopy(sparse_nested_table)
    inner_cards = outer_inner_cards[1]

    all_inner_table_assignments = itertools.product(*[range(c) for c in inner_cards])
    for outer_assign in sparse_nested_inner_dense_table.keys():
        for inner_assign in all_inner_table_assignments:
            if inner_assign not in sparse_nested_inner_dense_table[outer_assign]:
                sparse_nested_inner_dense_table[outer_assign][inner_assign] = default_value
    return sparse_nested_inner_dense_table


def _get_nested_sorted_probs(
        new_variables_order_outer, new_variables_order_inner, old_variable_order, old_assign_probs
):
    """
    Reorder variables to a new order (and sort assignments) and then convert the probs dictionary to a hierarchical
    dictionary with certain variables (or rather their corresponding assignments) in the outer dictionary and the rest
    in the inner dictionaries.

    :param new_variables_order_outer: The variables (and their order) for the outer scope.
    :type new_variables_order_outer: str list
    :param new_variables_order_inner: The variables (and their order) for the inner scope.
    :type new_variables_order_inner: str list
    :param old_variable_order: The initial variable order corresponding to old_assign_probs/
    :type old_variable_order: str list
    :param old_assign_probs: A dictionary of assignment and corresponding probabilities.
    :type old_assign_probs: dict

    Example:
        old_variable_order = [a, b]
        new_variables_order_outer = [b]

          a  b  c   P(a,b)     return:       b    a  c  P(b,a)
        {(0, 0, 0): pa0b0c0                {(0):{(0, 0): pa0b0,
         (0, 1, 0): pa0b1c0                      (1, 0): pa1b0}
         (1, 0, 1): pa1b0c1                 (1):{(0, 1): pa0b1,
         (1, 1, 1): pa1b1c1}                     (1, 1): pa1b1}}
    """
    new_variable_order = new_variables_order_outer + new_variables_order_inner
    new_order_indices = [new_variable_order.index(var) for var in old_variable_order]
    new_assign_probs = dict()
    for old_assign_i, old_prob_i in old_assign_probs.items():
        new_row_assignment = [None] * len(old_assign_i)
        for old_i, new_i in enumerate(new_order_indices):
            new_row_assignment[new_i] = old_assign_i[old_i]
        l1_assign = tuple(new_row_assignment[: len(new_variables_order_outer)])
        if l1_assign not in new_assign_probs:
            new_assign_probs[l1_assign] = dict()
        assign_l2 = tuple(new_row_assignment[len(new_variables_order_outer):])
        new_assign_probs[l1_assign][assign_l2] = old_prob_i
    return new_assign_probs, new_variable_order


def _same_scope_binary_operation(probs_table_a, pribs_table_b, func, default):
    """
    Apply a mathematical operation between the two factors with the same variable scope.
    NB: this function assumes that the variables corresponding to the keys in the two different dicts have the same
     order.

    :param dict probs_table_a: The probs dictionary for factor (typically sub factor) A.
    :param dict pribs_table_b: The probs dictionary for factor (typically sub factor) B.
    """
    result_common_sub_dict = dict()
    all_common_assignments = set([*probs_table_a.keys()] + [*pribs_table_b.keys()])
    new_default = func(default, default)
    for assignment in all_common_assignments:
        a_val = probs_table_a.get(assignment, default)
        b_val = pribs_table_b.get(assignment, default)
        r_val = func(a_val, b_val)
        if r_val != new_default:
            result_common_sub_dict[assignment] = r_val
    return result_common_sub_dict


def _flatten_ntd(ntd):
    """
    Flatten a two level nested table dictionary.
    """
    flattened_ntd = dict()
    for assign_outer in ntd.keys():
        for assign_inner in ntd[assign_outer].keys():
            flattened_ntd[assign_outer + assign_inner] = ntd[assign_outer][assign_inner]
    return flattened_ntd


def _apply_any_scope_binary_opp(ntd_a,
                                ntd_b,
                                full_sd_a_default_sub_table,
                                full_sd_b_default_sub_table,
                                symmetric_difference_combinations,
                                func,
                                default):
    """
    Apply a binary operation using the two nested tables, a list of all the outer assignments that need to be process
     and default inner sub tables for each. This function loops through the outer assignments in the
    symmetric_difference_combinations, and gets the inner sub-table for each nested distribution dictionary (if the
    assignment is not present, the respective default sub-tables are used). The binary operation applied using these
    sub-table pairs and the result, finally, forms a part of the resulting final table.

    :param dict ntd_a: The one nested dictionary representing the distribution.
    :param dict ntd_b: The other nested dictionary representing the distribution.
    :param dict full_sd_a_default_sub_table: The default sub-table to be used for missing outer entries in ntd_a
    :param dict full_sd_b_default_sub_table: The default sub-table to be used for missing outer entries in ntd_a
    :param iterable symmetric_difference_combinations: The list of outer scope assignments that need to be processed.
    :param func: The binary operation.
    :param default: The default value for the inner dictionaries
    :return: The resulting distribution dictionary.
    :rtype: dict
    """
    resulting_factor_table = dict()
    for sd_assign_a, sd_assign_b in symmetric_difference_combinations:
        joined_sd_assignment = sd_assign_a + sd_assign_b

        # TODO: Add special, pre-computed result if both 0.
        common_vars_subtable_a = ntd_a.get(sd_assign_a, full_sd_a_default_sub_table)
        common_vars_subtable_b = ntd_b.get(sd_assign_b, full_sd_b_default_sub_table)

        # TODO: extend _same_scope_binary_operation functionality to allow table completion given flag.
        subtable_result = _same_scope_binary_operation(
            common_vars_subtable_a, common_vars_subtable_b, func, default
        )
        if subtable_result:
            resulting_factor_table[joined_sd_assignment] = subtable_result
    return resulting_factor_table


def _any_scope_bin_opp_none(ntd_a, ntd_b, outer_inner_cards_a, outer_inner_cards_b, func, default):
    """
    Apply a binary operation between categorical tables that have the same, disjoint or overlapping variable scopes.
    This function assumes that no combination of default or non-default values is guaranteed to result in a default
    value.
    :param ntd_a: The one nested dictionary representing the distribution.
    :param ntd_b: The other nested dictionary representing the distribution.
    ::param outer_inner_cards_a: The cardinalities of the inner and outer variables respectively in ntd_a
     (i.e [[2],[2,3]]).
    :param outer_inner_cards_b: The cardinalities of the inner and outer variables respectively in ntd_b
     (i.e [[2],[2,3]]).
    :param default: The default value.
    :return: The resulting distribution dictionary.
    :rtype: dict
    """
    # we have to check everything - any combination of {value, default_value} could result in non-default value.
    ntd_a_sd_assignments = list(itertools.product(*[range(c) for c in outer_inner_cards_a[0]]))
    ntd_b_sd_assignments = list(itertools.product(*[range(c) for c in outer_inner_cards_b[0]]))

    full_sd_a_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_a[1], default)
    full_sd_b_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_b[1], default)
    symmetric_difference_combinations = itertools.product(*[ntd_a_sd_assignments, ntd_b_sd_assignments])
    # Having only dense default table is not good enough, as this will still not cause the missing default values in
    # the sub tables to be used to compute potentially none default result values. we therefore need to make these
    # sub tables dense.
    ntd_a = _make_inner_dense(ntd_a, outer_inner_cards_a, default)
    ntd_b = _make_inner_dense(ntd_b, outer_inner_cards_b, default)
    resulting_table = _apply_any_scope_binary_opp(ntd_a,
                                                  ntd_b,
                                                  full_sd_a_default_sub_table,
                                                  full_sd_b_default_sub_table,
                                                  symmetric_difference_combinations,
                                                  func,
                                                  default)
    return resulting_table


def _any_scope_bin_opp_both(ntd_a, ntd_b, outer_inner_cards_a, outer_inner_cards_b, func, default):
    """
    Apply a binary operation between categorical tables that have the same, disjoint or overlapping variable scopes.
    This function assumes that the result will be default only if both ntd_a and ntd_b has a default value.
    :param ntd_a: The one nested dictionary representing the distribution.
    :param ntd_b: The other nested dictionary representing the distribution.
    :param outer_inner_cards_a: The cardinalities of the inner and outer variables respectively in ntd_a
     (i.e [[2],[2,3]]).
    :param outer_inner_cards_b: The cardinalities of the inner and outer variables respectively in ntd_b
     (i.e [[2],[2,3]]).
    :param default: The default value.
    :return: The resulting distribution dictionary.
    :rtype: dict
    """
    # The result will only be default if the values in both ntd_a are default. This means
    # that we have to extend the assignments in the symmetric_difference_combinations to
    # those the full set of all possible assignments where either ntd_a is not default or
    # ntd_b is not default, because we need to calculate these results explicitly.
    # For example:
    # ntd_a:
    #    a b
    #    0 0  v
    #   (1 0) d
    # ntd_b:
    #    a c
    #    0 0  v
    #    1 0  v
    # The full set of resulting operations would be:
    # a b c
    # 0 0 0 vv
    # 0 0 1 vd
    # 0 1 0 dv
    # 0 1 1 dd = d
    # 1 0 0 dv
    # 1 0 1 dd = d
    # 1 1 0 dv
    # 1 1 1 dd = d
    # But because a:1 is not in ntd_a, this will not be calculated automatically and we therefore need to add this
    # entry. We do this by adding default sub tables below.
    outer_assignments_a = set(ntd_a.keys())
    outer_assignments_b = set(ntd_b.keys())
    ntd_a_sd_assignments = list(outer_assignments_a.union(outer_assignments_b))
    # TODO: replace this deepcopy with something faster
    ntd_b_sd_assignments = copy.deepcopy(ntd_a_sd_assignments)
    full_sd_a_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_a[1], default)
    full_sd_b_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_b[1], default)
    # Calculate the symmetric_difference_combinations
    # The symmetric_difference_combinations are the combinations of the sub-assignments that are not common to both
    # factors. Depending on the value of default_rules, these assignments could only include assignments actually
    # specified in the nested table dictionaries, or some or all of the missing assignments (if default values in
    # either) table does not necessarily guarantee default results.
    symmetric_difference_combinations = itertools.product(*[ntd_a_sd_assignments, ntd_b_sd_assignments])
    resulting_table = _apply_any_scope_binary_opp(ntd_a,
                                                  ntd_b,
                                                  full_sd_a_default_sub_table,
                                                  full_sd_b_default_sub_table,
                                                  symmetric_difference_combinations,
                                                  func,
                                                  default)
    return resulting_table


def _any_scope_binary_operation(
        ntd_a, outer_inner_cards_a, ntd_b, outer_inner_cards_b, func, default, default_rules="none"
):
    """
    Apply a binary operation between categorical tables that have the same, disjoint or overlapping variable scopes.

    :param dict ntd_a: The nested table dictionary for factor A.
    :param list outer_inner_cards_a: The cardinalities for the outer and inner variables for factor A (i.e [[2,2], [3]])
    :param dict ntd_b: The nested table dictionary for factor B.
    :param list outer_inner_cards_b: The cardinalities for the outer and inner variables for factor B (i.e [[2,2], [3]])
    :param float default: The default value for ntd_a and ntd_b (the values for the missing values).
    :param str default_rules: The rules for when a calculation results will result in a default value (optional).
        This can help speed up this function. The possible values are as follows:
            'any' : If either ntd_a or ntd_b has a default value, the result will be default.
            'both': Only if both ntd_a and ntd_b has a default value, the result will be default.
            'none': No combination of default or non-default values is guaranteed to result in a default value
        If this parameter is not specified, 'none' will be used to ensure correct, albeit slower computation.
    :returns: The nested table dictionary for the resulting factor.
    :rtype: dict
    """

    # TODO: this will not work for tables with the same scope, as this will mess up the nestedness(?).
    #  Seems like it works - confirm.

    ntd_a_sd_assignments = ntd_a.keys()
    ntd_b_sd_assignments = ntd_b.keys()
    # Note: sd: symmetric_difference

    full_sd_a_default_sub_table = None
    full_sd_b_default_sub_table = None

    if default_rules == "none":
        resulting_factor_table = _any_scope_bin_opp_none(ntd_a, ntd_b, outer_inner_cards_a, outer_inner_cards_b, func,
                                                         default)
    elif default_rules == "both":
        resulting_factor_table = _any_scope_bin_opp_both(ntd_a, ntd_b, outer_inner_cards_a, outer_inner_cards_b, func,
                                                         default)

    # TODO: This is strange - improve.
    # TODO: See if any performance gains are possible with 'left' and 'right' flags and add them if so.
    elif default_rules == "any":
        symmetric_difference_combinations = itertools.product(*[ntd_a_sd_assignments, ntd_b_sd_assignments])
        resulting_factor_table = _apply_any_scope_binary_opp(ntd_a,
                                                             ntd_b,
                                                             full_sd_a_default_sub_table,
                                                             full_sd_b_default_sub_table,
                                                             symmetric_difference_combinations,
                                                             func,
                                                             default)
    else:
        raise ValueError(f'Unsupported value for {default_rules} default_rules')

    return resulting_factor_table


def _fast_copy_probs_table(table):
    """
    Copy a dictionary representation of a probability table faster than the standard deepcopy.

    :param dict table: A dictionary with the tuples of ints as keys and floats as values.
    :return: The copied table.
    """
    table_copy = {tuple(assign): value for assign, value in table.items()}
    return table_copy


class SparseCategorical(Factor):
    """
    A class for instantiating sparse tables with log probabilities.
    """

    def __init__(self, var_names, cardinalities, log_probs_table=None, probs_table=None, default_log_prob=-np.inf):
        """
        Construct a SparseLogTable. Either log_probs_table or probs_table should be supplied.

        :param var_names: The variable names.
        :type var_names: str list
        :param cardinalities: The cardinalities of the variables (i.e, for three binrary variables: [2,2,2])
        :type cardinalities: int list
        :param log_probs_table: A dictionary with assignments (tuples) as keys and probabilities as values.
            Missing assignments are assumed to have zero probability.
        :type log_probs_table: dict
        :param probs_table: A dictionary with assignments (tuples) as keys and log probabilities as values.
            Missing assignments are assumed to have -inf log-probability (zero probability).
        :type probs_table: dict

        Example:

        .. code-block:: python

            >>> var_names = ['a','b']
            >>> probs_table = {(0,0):0.8,
            ...                (0,1):0.2,
            ...                (1,0):0.4,
            ...                (1,1):0.6}
            >>> cardinalities = [2,2]
            >>> table = SparseCategorical(log_probs_table=log_probs_table,
            >>>                           var_names=var_names,
            >>>                           cardinalities=cardinalities)
        """
        # TODO: add check that assignment lengths are consistent with var_names
        # TODO: add check that cardinalities are consistent with assignments
        super().__init__(var_names=var_names)
        if len(cardinalities) != len(var_names):
            raise ValueError("The cardinalities and var_names lists should be the same length.")
        if (log_probs_table is None) and (probs_table is None):
            raise ValueError("Either log_probs_table or probs_table must be specified")
        if log_probs_table is None:
            log_probs_table = {assignment: np.log(prob) for assignment, prob in probs_table.items()}
        self.log_probs_table = _fast_copy_probs_table(log_probs_table)
        self.var_cards = dict(zip(var_names, cardinalities))
        self.cardinalities = cardinalities
        self.default_log_prob = default_log_prob

    # TODO: Improve this to take missing assignments into account. Alternatively: add functionality to sparsify factor
    #  when probs turn to 0.
    # TODO: Add variable order sorting

    def _all_non_default_equal(self, factor, rtol, atol):
        """
        Check that all non default values in this factor are the same as the corresponding values in factor, where
        the two factors have the same variable scope.

        :param factor: The other factor
        :param float rtol: The relative tolerance to use for factor equality check.
        :param float atol: The absolute tolerance to use for factor equality check.
        :return: The result of the check.
        :rtype: bool
        """

        for assign, factor_log_prob in self.log_probs_table.items():

            if assign not in factor.log_probs_table:
                # TODO: make this `not is_close` and simplify this method
                if factor_log_prob != factor.default_log_prob:
                    return False
            else:
                self_log_prob = factor.log_probs_table[assign]
                if not np.isclose(self_log_prob, factor_log_prob, rtol=rtol, atol=atol):
                    return False
        return True

    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: SparseCategorical
        :param float rtol: The relative tolerance to use for factor equality check.
        :param float atol: The absolute tolerance to use for factor equality check.
        :return: The result of the comparison.
        :rtype: bool
        """
        factor_ = factor
        if not isinstance(factor_, SparseCategorical):
            raise TypeError(f"factor must be of SparseCategorical type but has type {type(factor)}")

        if set(self.var_names) != set(factor_.var_names):
            return False

        # var sets are the same
        if self.var_names != factor.var_names:
            factor_ = factor.reorder(self.var_names)
        # factors now have same variable order

        # everywhere that self has non default values, factor has the same values.
        if not self._all_non_default_equal(factor_, rtol=rtol, atol=atol):
            return False
        # TODO: improve efficiency here (there could be a lot of duplication with the above loop)
        #   Check the values for every non-default assignment of factor
        if not factor_._all_non_default_equal(self, rtol=rtol, atol=atol):
            return False

        # If all possible assignments have not been checked - check that the default values are the same
        if not self._is_dense() and not factor._is_dense():
            if self.default_log_prob != factor.default_log_prob:
                return False

        return True

    def copy(self):
        """
        Make a copy of this factor.

        :return: The copy of this factor.
        :rtype: SparseCategorical
        """
        return SparseCategorical(
            var_names=self.var_names.copy(),
            log_probs_table=_fast_copy_probs_table(self.log_probs_table),
            cardinalities=copy.deepcopy(self.cardinalities),
            default_log_prob=self.default_log_prob,
        )

    def marginalize(self, vrs, keep=True):
        """
        Sum out variables from this factor.

        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: The resulting factor.
        :rtype: SparseCategorical
        """

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_sum_out = [v for v in self.var_names if v not in vars_to_keep]
        nested_table, _ = _get_nested_sorted_probs(
            new_variables_order_outer=vars_to_keep,
            new_variables_order_inner=vars_to_sum_out,
            old_variable_order=self.var_names,
            old_assign_probs=self.log_probs_table,
        )
        result_table = dict()
        for l1_assign, log_probs_table in nested_table.items():
            prob = special.logsumexp(list(log_probs_table.values()))
            result_table[l1_assign] = prob

        result_var_cards = copy.deepcopy(self.var_cards)
        for var in vars_to_sum_out:
            del result_var_cards[var]
        cardinalities = list(result_var_cards.values())
        return SparseCategorical(
            var_names=vars_to_keep, log_probs_table=result_table, cardinalities=cardinalities
        )

    def reduce(self, vrs, values):
        """
        Observe variables to have certain values and return reduced table.

        :param vrs: (list) The variables.
        :param values: (tuple or list) Their values
        :return: The resulting factor.
        :rtype: SparseCategorical
        """

        vars_unobserved = [var_name for var_name in self.var_names if var_name not in vrs]
        nested_table, _ = _get_nested_sorted_probs(
            new_variables_order_outer=vrs,
            new_variables_order_inner=vars_unobserved,
            old_variable_order=self.var_names,
            old_assign_probs=self.log_probs_table,
        )
        lp_table = nested_table[tuple(values)]
        result_var_cards = copy.deepcopy(self.var_cards)
        for var in vrs:
            del result_var_cards[var]

        cards = list(result_var_cards.values())
        var_names = vars_unobserved
        return SparseCategorical(var_names=var_names, log_probs_table=lp_table, cardinalities=cards)

    def _assert_consistent_cardinalities(self, factor):
        """
        Assert that the variable cardinalities are consistent between two factors.

        :param SparseCategorical factor: The factor to compare with.
        """
        for var in self.var_names:
            if var in factor.var_cards:
                error_msg = f"Error: inconsistent variable cardinalities: {factor.var_cards}, {self.var_cards}"
                assert self.var_cards[var] == factor.var_cards[var], error_msg

    def multiply(self, factor):
        """
        Multiply this factor with another factor and return the result.

        :param factor: The factor to multiply with.
        :type factor: SparseCategorical
        :return: The factor product.
        :rtype: SparseCategorical
        """
        if not isinstance(factor, SparseCategorical):
            raise TypeError(f"factor must be of SparseCategorical type but has type {type(factor)}")
        return self._apply_binary_operator(factor, operator.add, default_rules="any")

    def cancel(self, factor):
        """
        Almost like divide, but with a special rule that ensures that division of zeros by zeros results in zeros.

        :param factor: The factor to divide by.
        :type factor: SparseCategorical
        :return: The factor quotient.
        :rtype: SparseCategorical
        """

        def special_divide(val_a, val_b):
            if (val_a == -np.inf) and (val_b == -np.inf):
                return -np.inf
            return val_a - val_b

        return self._apply_binary_operator(factor, special_divide, default_rules="any")

    def divide(self, factor):
        """
        Divide this factor by another factor and return the result.

        :param factor: The factor to divide by.
        :type factor: SparseCategorical
        :return: The factor quotient.
        :rtype: SparseCategorical
        """
        return self._apply_binary_operator(factor, operator.sub, default_rules="none")

    def argmax(self):
        """
        Get the Categorical assignment (vector value) that maximises the factor potential.

        :return: The argmax assignment.
        :rtype: int list
        """
        return max(self.log_probs_table.items(), key=operator.itemgetter(1))[0]

    def _apply_to_probs(self, func, include_assignment=False):
        """
        Apply a function to the log probs of the factor.

        :param func: The function to apply to the log probs in this factor.
        :param include_assignment: Whether or not to pass the assignment to the function as well
            (along with the log probs).
        """
        for assign, prob in self.log_probs_table.items():
            if include_assignment:
                self.log_probs_table[assign] = func(prob, assign)
            else:
                self.log_probs_table[assign] = func(prob)

    def _apply_binary_operator(self, factor, operator_function, default_rules="none"):
        """
        Apply a binary operator function f(self.factor, factor) and return the result

        :param factor: The other factor to use in the binary operation.
        :type factor: SparseCategorical
        :param default_rules: The rules for when a calculation results will result in a default value (optional).
            This can help speed up this function. The possible values are as follows:
                'left' : If ntd_a has a default value, the result will always be default.
                'right' : If ntd_b has a default value, the result will always be default.
                'any' : If either ntd_a or ntd_b has a default value, the result will always be default.
                'both': Only if both ntd_a and ntd_b has a default value, the result will always be default.
                'none': No combination of default or non-default values is guarenteed to result in a default value
        If this parameter is not specified, 'none' will be used to ensure correct, albeit slower computation.
        :return: The resulting factor.
        :rtype: SparseCategorical
        """
        # pylint: disable=too-many-locals
        if not isinstance(factor, SparseCategorical):
            raise TypeError(f"factor must be of SparseCategorical type but has type {type(factor)}")
        self._assert_consistent_cardinalities(factor)
        intersection_vars = list(set(self.var_names).intersection(set(factor.var_names)))
        intersection_vars = sorted(intersection_vars)

        remaining_a_vars = list(set(self.var_names) - set(intersection_vars))
        ntd_a, _ = _get_nested_sorted_probs(
            new_variables_order_outer=remaining_a_vars,
            new_variables_order_inner=intersection_vars,
            old_variable_order=self.var_names,
            old_assign_probs=self.log_probs_table,
        )

        remaining_b_vars = list(set(factor.var_names) - set(intersection_vars))
        ntd_b, _ = _get_nested_sorted_probs(
            new_variables_order_outer=remaining_b_vars,
            new_variables_order_inner=intersection_vars,
            old_variable_order=factor.var_names,
            old_assign_probs=factor.log_probs_table,
        )

        # TODO: Add this functionality
        if self.default_log_prob != factor.default_log_prob:
            error_msg = (
                "Cases where self.default_value and factor.default_value differ are not yet supported."
            )
            raise NotImplementedError(error_msg)
        default_log_prob = self.default_log_prob

        def vars_to_cards(factor, var_names):
            return [factor.var_cards[v] for v in var_names]

        outer_inner_cards_a = [vars_to_cards(self, remaining_a_vars), vars_to_cards(self, intersection_vars)]
        outer_inner_cards_b = [
            vars_to_cards(factor, remaining_b_vars),
            vars_to_cards(factor, intersection_vars),
        ]

        result_ntd = _any_scope_binary_operation(
            ntd_a,
            outer_inner_cards_a,
            ntd_b,
            outer_inner_cards_b,
            operator_function,
            default_log_prob,
            default_rules=default_rules,
        )
        flattened_result_table = _flatten_ntd(result_ntd)
        result_var_names = remaining_a_vars + remaining_b_vars + intersection_vars
        result_var_cards = {**self.var_cards, **factor.var_cards}
        result_cardinalities = [result_var_cards[v] for v in result_var_names]
        result_default_log_prob = operator_function(default_log_prob, default_log_prob)
        return SparseCategorical(
            var_names=result_var_names,
            log_probs_table=flattened_result_table,
            cardinalities=result_cardinalities,
            default_log_prob=result_default_log_prob,
        )

    def normalize(self):
        """
        Normalize the factor.

        :return: The normalized factor.
        :rtype: SparseCategorical
        """

        factor_copy = self.copy()
        logz = special.logsumexp(list(factor_copy.log_probs_table.values()))
        for assign, prob in factor_copy.log_probs_table.items():
            factor_copy.log_probs_table[assign] = prob - logz
        return factor_copy

    def _is_dense(self):
        """
        Check if all factor assignments have non-default values.

        :return: The result of the check.
        :rtype: Bool
        """
        num_assignments = len(self.log_probs_table)
        max_possible_assignments = np.prod(self.cardinalities)
        if num_assignments == max_possible_assignments:
            return True
        return False

    @staticmethod
    def _raw_kld(log_p, log_q):
        """
        Get the raw numerically calculated kld (which could result in numerical errors causing negative KLDs).

        :param log_p: The log_p tensor
        :param log_q:
        :return: The KL-divergence
        :rtype: float
        """

        def kld_from_log_elements(log_pi, log_qi):
            if log_pi == -np.inf:
                # lim_{p->0} p*log(p/q) = 0
                # lim_{p->0} p*log(p/q) = 0, with q = p
                # So if p_i = 0, kld_i = 0
                return 0.0
            kld_i = np.exp(log_pi) * (log_pi - log_qi)
            return kld_i

        kld_factor = SparseCategorical._apply_binary_operator(
            log_p, log_q, kld_from_log_elements, default_rules="none"
        )  # TODO: check this

        klds = list(kld_factor.log_probs_table.values())
        kld = np.sum(klds)
        return kld

    def kl_divergence(self, factor, normalize_factor=True):
        """
        Get the KL-divergence D_KL(P || Q) = D_KL(self||factor) between a normalized version of this factor and another factor.
        Reference https://infoscience.epfl.ch/record/174055/files/durrieuThiranKelly_kldiv_icassp2012_R1.pdf, page 1.

        :param factor: The other factor
        :type factor: Gaussian
        :param normalize_factor: Whether or not to normalize the other factor before computing the KL-divergence.
        :type normalize_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        log_p = self.normalize()
        log_q = factor
        if normalize_factor:
            log_q = factor.normalize()

        kld = self._raw_kld(log_p, log_q)

        if kld < 0.0:
            if np.isclose(kld, 0.0, atol=1e-4):
                #  this is fine (numerical error)
                return 0.0
            kld_msg_details = "Categorical:\n" + log_p.__repr__() + "\nfactor:" + log_q.__repr__()
            raise ValueError(f"Negative KLD: {kld}. Details:\n{kld_msg_details}")
        return kld

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.

        :return: The KL divergence.
        :rtype: float
        """
        # make uniform copy
        uniform_factor = self.copy()
        cards = list(uniform_factor.var_cards.values())
        uniform_log_prob = -np.log(np.product(cards))
        uniform_factor._apply_to_probs(lambda x: uniform_log_prob)
        kld = self.kl_divergence(uniform_factor, normalize_factor=False)
        return kld

    def potential(self, vrs, assignment):
        """
        Get the value of the factor for a specific assignment.

        :param assignment: The assignment
        :return: The value
        """
        assert set(vrs) == set(self.var_names), "variables (vrs) do not match factor variables."
        vrs_to_var_names_indices = [self.var_names.index(v) for v in vrs]
        var_names_order_assignments = tuple([assignment[i] for i in vrs_to_var_names_indices])
        return np.exp(self.log_probs_table[var_names_order_assignments])

    def _to_df(self):
        """
        Convert the factor to a dataframe representation.

        :return: The dataframe representation.
        :rtype: pandas.DataFrame
        """
        log_probs_table = self.log_probs_table
        var_names = self.var_names
        factor_df = pd.DataFrame.from_dict(log_probs_table.items()).rename(columns={0: "assignment", 1: "log_prob"})
        factor_df[var_names] = pd.DataFrame(factor_df["assignment"].to_list())
        factor_df.drop(columns=["assignment"], inplace=True)
        # Correct column order
        var_cols = [c for c in factor_df.columns if c != "log_prob"]
        return factor_df[var_cols + ["log_prob"]]

    def show(self):
        """
        Print the factor's string representation.
        """
        print(self.__repr__())

    def __repr__(self):
        """
        Get the string representation for the factor.

        :return: The representation string
        :rtype: str
        """
        tabbed_spaced_var_names = "\t".join(self.var_names) + "\tprob\n"
        repr_str = tabbed_spaced_var_names
        spacings = ['\t' * _factor_utils.tabs_to_cover_string(var) for var in self.var_names] + ['\n']
        for assignment, log_prob in self.log_probs_table.items():
            prob = np.exp(log_prob)
            line = _factor_utils.space_assignments_and_probs(assignment, prob, spacings)
            repr_str += line
        return repr_str

    def reorder(self, new_var_names_order):
        """
        Reorder categorical table variables to a new order and reorder the associated probabilities
        accordingly.

        :param new_var_names_order: The new variable order.
        :type new_var_names_order: str list
        :return: The factor with new order.

        Example:
            old_variable_order = [a, b]
            new_variable_order = [b, a]

            a b P(a,b)  return    b a P(a,b)
            0 0  pa0b0            0 0  pa0b0
            0 1  pa0b1            0 1  pa1b0
            1 0  pa1b0            1 0  pa0b1
            1 1  pa1b1            1 1  pa1b1
        """

        new_order_indices = [self.var_names.index(var) for var in new_var_names_order]
        new_log_probs_table = dict()
        for assignment, value in self.log_probs_table.items():
            reordered_assignment = tuple(assignment[i] for i in new_order_indices)
            new_log_probs_table[reordered_assignment] = value
        reordered_cardinalities = [self.cardinalities[i] for i in new_order_indices]

        return SparseCategorical(
            var_names=new_var_names_order,
            log_probs_table=new_log_probs_table,
            cardinalities=reordered_cardinalities,
        )


class SparseCategoricalTemplate(FactorTemplate):
    """
    A class for specifying sparse categorical factor templates and creating categorical factors from these templates.
    """

    def __init__(self, probs_table, cardinalities, var_templates=None):
        """
        Create a Categorical factor template.

        :param probs_table: The log_probs_table that specifies the assignments and values for the template.
        :type probs_table: tuple:float dict
        :param cardinalities: The cardinalities of the variables (i.e, for three binrary variables: [2,2,2])
        :type cardinalities: int list
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        Example:

        .. code-block:: python

            >>> var_templates = ['a_{i}','b_{i}']
            >>> probs_table = {(0,0):0.8,
            ...                (0,1):0.2,
            ...                (1,0):0.4,
            ...                (1,1):0.6}
            >>> cardinalities = [2,2]
            >>> categorical_template = SparseCategoricalTemplate(log_probs_table=log_probs_table,
            ...                                                  var_names=var_names,
            ...                                                  cardinalities=cardinalities)
            >>> categorical_template.make_factor(format_dict={'i':0})

        """
        # TODO: Complete and improve docstring.
        super().__init__(var_templates=var_templates)
        self.probs_table = _fast_copy_probs_table(probs_table)
        self.cardinalities = cardinalities

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :return: The instantiated factor.
        :rtype: SparseCategorical
        """
        if (self._var_templates is None) and (var_names is None):
            raise ValueError(
                "var_names need to be supplied to make a factor from SparseCategoricalTemplate without var_templates"
            )
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        cards = list(self.cardinalities)
        factor = SparseCategorical(probs_table=self.probs_table, var_names=var_names, cardinalities=cards)
        return factor



"""
A module for instantiating and performing operations on multivariate Gaussian mixture distributions.
"""

# pylint: disable=cyclic-import
# pylint: disable=protected-access

# Standard imports
import cmath

# Third-party imports
import numpy as np
import numdifftools as nd
from matplotlib import pyplot as plt
from scipy.optimize import minimize
from scipy import special

# Local imports
from veroku.factors._factor import Factor
from veroku.factors import _factor_utils
import veroku.factors.gaussian as gauss
from veroku._constants import DEFAULT_FACTOR_RTOL, DEFAULT_FACTOR_ATOL

# TODO: confirm that the different GaussianMixture divide methods have sufficient stability and accuracy in their
#  approximations.
# TODO: Add tests for the divide methods.
# TODO: move to factors (non-experimental) once the divide methods have been checked and tested properly.


class GaussianMixture(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    def __init__(self, factors, cancel_method=0):
        """
        A GaussianMixture constructor.

        :param factors: a list of Gaussian type objects with the same dimensionality.
        :type factors: Gaussian list
        :param cancel_method: The method of performing approximate division.
                              This is only applicable if the factor is a Gaussian mixture with more than one component.
                              0: moment match denominator to single Gaussian and divide
                              1: approximate modes of quotient as Gaussians
                              2: Use complex moment matching on inverse mixture sets (s_im) and approximate the inverse
                                 of each s_im as a Gaussian
        :type cancel_method: int
        """
        assert factors, "Error: empty list passed to constructor."
        self.cancel_method = cancel_method
        self.components = [gaussian.copy() for gaussian in factors]
        self.num_components = len(factors)

        var_names0 = factors[0].var_names

        for component in self.components:
            if var_names0 != component.var_names:
                raise ValueError("inconsistent var_names in list of Gaussians.")
        super().__init__(var_names=var_names0)

    def equals(self, factor, rtol=DEFAULT_FACTOR_RTOL, atol=DEFAULT_FACTOR_ATOL):
        """
        Check if this factor is the same as another factor.

        :param factor: The factor to compare with.
        :type factor: GaussianMixture
        :param float rtol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :param float atol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :return: Result of equals comparison between self and gaussian
        rtype: bool
        """
        # TODO: consider adding type error here rather
        if not isinstance(factor, GaussianMixture):
            raise TypeError(f"factor must be of GaussianMixture type but has type {type(factor)}")
        if factor.num_components != self.num_components:
            return False
        for i in range(self.num_components):
            found_corresponding_factor = False
            for j in range(i, self.num_components):
                if self.get_component(i).equals(factor.get_component(j)):
                    found_corresponding_factor = True
            if not found_corresponding_factor:
                return False
        return True

    def get_component(self, index):
        """
        Get the Gaussian component at an index.

        :param index: The index of teh component to return.
        :type index: int
        :return: The component at the given index.
        :rtype: Gaussian
        """
        return self.components[index]

    def copy(self):
        """
        Make a copy of this Gaussian mixture.

        :return: The copied GaussianMixture.
        :rtype: GaussianMixture
        """
        component_copies = []
        for comp in self.components:
            component_copies.append(comp.copy())
        return GaussianMixture(component_copies)

    def multiply(self, factor):
        """
        Multiply this GaussianMixture with another factor.

        :param factor: The factor to multiply with.
        :type factor: Gaussian or Gaussian Mixture
        :return: The factor product.
        :rtype: GaussianMixture
        """
        new_components = []
        if isinstance(factor, gauss.Gaussian):
            for comp in self.components:
                new_components.append(comp.multiply(factor))
        elif isinstance(factor, GaussianMixture):
            for comp_ai in self.components:
                for comp_bi in factor.components:
                    new_components.append(comp_ai.multiply(comp_bi))
        else:
            raise TypeError("unsupported factor type.")
        return GaussianMixture(new_components)

    def divide(self, factor):
        """
        Divide this GaussianMixture by another factor.

        :param factor: The factor divide by.
        :type factor: Gaussian or Gaussian Mixture
        :return: The resulting factor quotient (approximate in the case of where both the numerator and denominator are
        GaussianMixtures with more than one component).
        :rtype: GaussianMixture
        """
        if isinstance(factor, gauss.Gaussian):
            single_gaussian = factor
        elif isinstance(factor, GaussianMixture):
            if factor.num_components == 1:
                single_gaussian = factor.get_component(index=0)
            else:
                if self.cancel_method == 0:
                    single_gaussian = factor.moment_match_to_single_gaussian()
                if self.cancel_method == 1:
                    return GaussianMixture._gm_division_m1(self, factor)
                if self.cancel_method == 2:
                    return GaussianMixture._gm_division_m2(self, factor)
        else:
            raise TypeError("unsupported factor type.")
        new_components = []
        for comp in self.components:
            new_components.append(comp.divide(single_gaussian))
        return GaussianMixture(new_components)

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this Gaussian mixture and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :type vrs: str list
        :param values: the values of the observed variables (list or vector-like object)
        :type values: vector-like
        :return: the observation reduced factor.
        :rtype: GaussianMixture
        """
        new_components = []
        for comp in self.components:
            new_components.append(comp.reduce(vrs, values))
        return GaussianMixture(new_components)

    def marginalize(self, vrs, keep=True):
        """
        Integrate out variables from this Gaussian mixture.

        :param vrs: A subset of variables in the factor's scope.
        :type vrs: str list
        :param keep: Whether to keep or sum out vrs.
        :type keep: bool
        :return: the resulting marginal factor.
        :rtype: GaussianMixture
        """
        new_components = []
        for comp in self.components:
            new_components.append(comp.marginalize(vrs, keep))
        return GaussianMixture(new_components)

    def distance_from_vacuous(self):
        """
        NOTE: Not Implemented yet.
        Get the Kullback-Leibler (KL) divergence between the message factor and a vacuous (uniform) version of it.

        :return: The KL-divergence
        """
        raise NotImplementedError('This function has not been implemented yet.')

    def kl_divergence(self, factor):
        """
        NOTE: Not Implemented yet.
        Get the KL-divergence D_KL(self || factor) between a normalized version of this factor and another factor.

        :param factor: The other factor
        :type factor: GaussianMixture
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        raise NotImplementedError('This function has not been implemented yet.')

    def log_potential(self, x_val):
        """
        Get the log of the value of the Gaussian mixture potential at X.

        :param x_val: The point to evaluate the GaussianMixture at.
        :type x_val: vector-like
        :return: log of the value of the GaussianMixture potential at x.
        :rtype: float
        """
        log_potentials = []
        for comp in self.components:
            log_potentials.append(comp.log_potential(x_val))
        total_log_potx = special.logsumexp(log_potentials)
        return total_log_potx

    def potential(self, x_val):
        """
        Get the value of the Gaussian mixture potential at x_val.

        :param x_val: The point to evaluate the GaussianMixture at.
        :type x_val: vector-like
        :return: log of the value of the GaussianMixture potential at x_val.
        :rtype: float
        """
        total_potx = 0.0
        for comp in self.components:
            total_potx += comp.potential(x_val)
        return total_potx

    def _get_means(self):
        """
        Get the means of the Gaussian components.

        :return: the mean vectors
        """
        means = []
        for comp in self.components:
            means.append(comp.get_mean())
        return means

    def _get_covs(self):
        """
        Get the covariance matrices of the Gaussian components.

        :return: the covariance matrices
        :rtype: np.ndarray list
        """
        covs = []
        for comp in self.components:
            covs.append(comp.get_cov())
        return covs

    def _get_log_weights(self):
        """
        Get the log weights of the Gaussian mixture components.

        :return: the log weights
        :rtype: float
        """
        log_weights = []
        for comp in self.components:
            log_weights.append(comp.get_log_weight())
        return log_weights

    def _get_weights(self):
        """
        Get the weights of the Gaussian mixture components.

        :return: the log weights
        :rtype: float list
        """
        weights = []
        for comp in self.components:
            weights.append(comp.get_weight())
        return weights

    def get_log_weight(self):
        """
        Get the total log weight of the Gaussian mixture.

        :return: The log weight
        :rtype: float
        """
        return special.logsumexp(self._get_log_weights())

    def normalize(self):
        """
        Normalize the factor.

        :return: The normalized factor.
        :rtype: GaussianMixture
        """
        unnormalized_log_weight = self.get_log_weight()
        new_components = []
        for comp in self.components:
            comp_copy = comp.copy()
            comp_copy._add_log_weight(-1.0 * unnormalized_log_weight)
            new_components.append(comp_copy)
        return GaussianMixture(new_components)

    @property
    def is_vacuous(self):
        """
        Check if a Gaussian mixture distribution contains no information. This is the case when the K matrices of all
        components are zero matrices.

        :return: The result of the check.
        :rtype: Bool
        """
        # TODO: see how this is used. Should this be true if there is one vacuous component? Or only if all components
        #  are vacuous? Maybe make a contains vacuous function as well.
        for comp in self.components:
            if not comp._is_vacuous:
                return False
        return True

    def sample(self, num_samples):
        """
        Sample from this Gaussian mixture

        :param num_samples: The number of sample to draw.
        :type num_samples: int
        :return: The samples
        :rtype: float list
        """
        weights = self._get_weights()
        component_choice_samples = np.random.choice(
            range(len(weights)), size=num_samples, p=weights / np.sum(weights)
        )
        samples = []
        for comp_index in component_choice_samples:
            samples.append(self.components[comp_index].sample(1)[0])
        return np.array(samples)

    def _get_sensible_xlim(self):
        """
        Helper function for plot function to get the x limits that contain the majority of the Gaussian mass.

        :return: The x limits.
        :rtype: float list
        """
        x_lower_candidates = []
        x_upper_candidates = []
        for comp in self.components:
            stddev = np.sqrt(comp.get_cov()[0, 0])
            x_lower_candidates.append(comp.get_mean()[0, 0] - 3 * stddev)
            x_upper_candidates.append(comp.get_mean()[0, 0] + 3 * stddev)
        x_lower = min(x_lower_candidates)
        x_upper = max(x_upper_candidates)
        return [x_lower, x_upper]

    def plot(self, log=False, xlim=None, ylim=None, show_individual_components=False):
        """
        Plot the Gaussian mixture potential function (only for 1d and 2d functions).

        :param log: if this is True, the log-potential will be plotted
        :type log: bool
        :param xlim: the x limits to plot the function over (optional)
        :type xlim: 2 element float list
        :param ylim: the y limits to plot the function over (optional and only used in 2d case)
        :type ylim: 2 element float list
        """
        if self.dim == 1:
            if xlim is None:
                xlim = self._get_sensible_xlim()
            if xlim is not None:
                x_lower = xlim[0]
                x_upper = xlim[1]
            num_points = 200
            x_series = np.linspace(x_lower, x_upper, num_points)
            total_potx = np.zeros(num_points)
            for comp in self.components:
                if log:
                    potx = np.array([comp.log_potential(xi) for xi in x_series])
                else:
                    potx = np.array([comp.potential(xi) for xi in x_series])
                if show_individual_components:
                    plt.plot(x_series, potx)
                total_potx += potx
            plt.plot(x_series, total_potx)
        elif self.dim == 2:
            self._plot_2d(log=log, xlim=xlim, ylim=ylim)
        else:
            raise NotImplementedError("Plotting not implemented for dim!=1.")

    def show(self):
        """
        Print the parameters of the Gaussian mixture distribution
        """
        for i, comp in enumerate(self.components):
            print("component ", i, "/", len(self.components))
            comp.show()

    def moment_match_to_single_gaussian(self):
        """
        Calculate the mean and covariance of the Gaussian mixture and return a Gaussian with these parameters as an
        approximation of the Gaussian Mixture.

        :return: The Gaussian approximation.
        :rtype: Gaussian
        """
        new_mean = np.zeros([self.dim, 1])
        log_weights = []
        for comp in self.components:
            weight_mean = comp.get_weight() * comp.get_mean()
            new_mean += weight_mean
            log_weights.append(comp.get_log_weight())
        log_sum_weights = special.logsumexp(log_weights)
        sum_weights = np.exp(log_sum_weights)
        new_mean = new_mean / sum_weights
        new_cov = np.zeros([self.dim, self.dim])
        for comp in self.components:
            udud_t = comp.get_weight() * (comp.get_mean() - new_mean).dot(
                (comp.get_mean() - new_mean).transpose()
            )
            new_cov += comp.get_weight() * comp.get_cov() + udud_t
        new_cov = new_cov / sum_weights
        return gauss.Gaussian(
            cov=new_cov, mean=new_mean, log_weight=log_sum_weights, var_names=self.components[0].var_names
        )

    def argmax(self):
        """
        Find the input vector that maximises the potential function of the Gaussian mixture.

        :return: The argmax assignment.
        :rtype: numpy.ndarray
        """
        global_maximum_potential = -float("inf")
        global_argmax = None
        success = False

        def neg_gmm_pot(x_val):
            return -1.0 * self.potential(x_val)

        for comp in self.components:
            res = minimize(neg_gmm_pot, x0=comp.get_mean(), method="BFGS", options={"disp": False})
            x_local_max = res.x
            if res.success:
                success = True
                local_maximum_potential = self.potential(x_local_max)
                if local_maximum_potential > global_maximum_potential:
                    global_maximum_potential = local_maximum_potential
                    global_argmax = x_local_max
        if not success:
            raise Exception("could not find optimum")
        return global_argmax

    def _argmin(self):
        """
        Find the input vector that minimises the potential function of the Gaussian mixture with negative definite
        precision matrices.

        :return: The point where the function has a global minimum.
        :rtype: np.ndarray
        """
        global_minimum_potential = float("inf")
        global_argmin = None
        success = False

        def gmm_pot(x_vals):
            return self.potential(x_vals)

        for comp in self.components:
            res = minimize(gmm_pot, x0=comp.get_mean(), method="BFGS", options={"disp": False})
            x_local_min = res.x
            if res.success:
                success = True
                local_minimum_potential = self.potential(x_local_min)
                if local_minimum_potential < global_minimum_potential:
                    global_minimum_potential = local_minimum_potential
                    global_argmin = x_local_min
        if not success:
            raise Exception("could not find optimum")
        return global_argmin

    def _moment_match_complex_gaussian(self):
        """
        Calculate the mean and covariance of the Gaussian mixture and return a Gaussian
        with these parameters a Gaussian approximation of the Gaussian Mixture.

        :return: The Gaussian approximation.
        :rtype: Gaussian
        """
        # TODO: check if the normal moment matching function can be replaced with this one.
        new_mean = np.zeros([self.dim, 1]).astype(complex)
        weights = []
        for comp in self.components:
            c_weight = comp._get_complex_weight()
            c_mean = np.linalg.inv(comp.get_prec()).dot(comp.get_h()).astype(complex)
            new_mean += c_weight * c_mean
            weights.append(c_weight)
        sum_weights = sum(weights)
        new_mean = new_mean / sum_weights
        new_cov = np.zeros([self.dim, self.dim]).astype(complex)
        for comp in self.components:
            c_weight = comp._get_complex_weight()
            c_mean = np.linalg.inv(comp.get_prec()).dot(comp.get_h()).astype(complex)
            udud_t = c_weight * (c_mean - new_mean).dot((c_mean - new_mean).transpose())
            c_cov = np.linalg.inv(comp.get_prec()).astype(complex)
            new_cov += c_weight * c_cov + udud_t
        new_cov = new_cov / sum_weights

        new_log_weight = np.log(sum_weights)

        new_prec, new_h_vec, new_g = GaussianMixture._complex_cov_form_to_real_canform(
            new_cov, new_mean, new_log_weight
        )
        return gauss.Gaussian(prec=new_prec, h_vec=new_h_vec, g_val=new_g, var_names=self.components[0].var_names)

    @staticmethod
    def _complex_cov_form_to_real_canform(cov, mean, log_weight):
        """
        A helper function for _moment_match_complex_gaussian.

        :param cov: the (possibly complex) covariance matrix
        :param mean: the (possibly complex) covariance matrix
        :param log_weight: the (possibly complex) log weight
        :return: the real parts of the converted canonical parameters (prec, h_vec, g_val)
        """
        prec = np.linalg.inv(cov)
        h_vec = prec.dot(mean)
        ut_k_u = ((mean.transpose()).dot(prec)).dot(mean)
        log_det_term = cmath.log(np.linalg.det(2.0 * np.pi * cov))
        g_val = abs(log_weight - 0.5 * ut_k_u - 0.5 * log_det_term)
        return prec.real, h_vec.real, g_val.real

    @staticmethod
    def _get_inverse_gaussians(gaussian_mixture_a, gaussian_mixture_b):
        """
        A helper function for _gm_division_m1. Returns the inverse components.

        :param gaussian_mixture_a: The numerator mixture
        :param gaussian_mixture_b: The denominator mixture
        :return: The inverse Gaussian components and the mode locations of the quotient:
                 gaussian_mixture_a/gaussian_mixture_b

        Example:
            if gaussian_mixture_a = Ga1 + Ga2
            and gaussian_mixture_b = Gb1 + Gb2
            Then return Gb1/Ga1 + Gb1/Ga2 + Gb2/Ga1 + Gb2/Ga2
        """
        minimum_locations = []
        for g_a in gaussian_mixture_a.components:
            inverse_components = []
            inverse_gaussian_mixtures = []
            for g_b in gaussian_mixture_b.components:
                inverse_components.append(g_b.divide(g_a))
            inverse_gaussian_mixture = GaussianMixture(inverse_components)
            inverse_gaussian_mixtures.append(inverse_gaussian_mixture)

            minimum_loc = inverse_gaussian_mixture._argmin()
            minimum_locations.append(minimum_loc)

        def neg_gm_log_quotient_potential(x_val):
            """
            Get the negative of the log value of the Gaussian mixture quotient (gma/gmb)
            """
            potential = -1.0 * (gaussian_mixture_a.log_potential(x_val) - gaussian_mixture_b.log_potential(x_val))
            return potential

        mode_locations = []
        for minimum_loc in minimum_locations:
            res = minimize(
                neg_gm_log_quotient_potential, x0=minimum_loc, method="BFGS", options={"disp": False}
            )
            x_local_min = res.x
            if res.success:
                mode_locations.append(x_local_min)
        unique_mode_locations = _factor_utils.remove_duplicate_values(mode_locations, tol=1e-3)
        return inverse_gaussian_mixtures, unique_mode_locations

    @staticmethod
    def _gm_division_m1(gaussian_mixture_a, gaussian_mixture_b):
        """
        Method 1 for approximating the quotient (gma/gmb) of two Gaussian mixtures as a Gaussian mixture.
        This is an implementation of the method described in 'A Probabilistic Graphical Model Approach to Multiple
        Object Tracking' (section 8.2) The Gaussian mixture quotient is optimised from a set of starting points in order
        to find modes. Laplaces method is then used at these modes to approximate the mode as a Gaussian. The quotient
        is then approximated as a sum of these Gaussians.

        :param GaussianMixture gma: The numerator factor
        :param GaussianMixture gmb: The denominator factor
        :returns: an approximation of the quotient function as a Gaussian mixture
        :rtype: GaussianMixture
        """

        resulting_gaussian_components = []
        inverse_gaussian_mixtures, unique_mode_locations = GaussianMixture._get_inverse_gaussians(
            gaussian_mixture_a, gaussian_mixture_b
        )
        inv_gm = GaussianMixture(inverse_gaussian_mixtures)

        for mode_location in unique_mode_locations:
            prec_i = nd.Hessian(inv_gm.log_potential)(mode_location)
            mean_i = _factor_utils.make_column_vector(mode_location)
            h_i = prec_i.dot(mean_i)

            ut_k_u = ((mean_i.transpose()).dot(prec_i)).dot(mean_i)
            log_weight_i = -0.5 * np.log(np.linalg.det(prec_i / (2.0 * np.pi))) * inv_gm.log_potential(mean_i)
            g_i = log_weight_i - 0.5 * ut_k_u + 0.5 * np.log(np.linalg.det(prec_i / (2.0 * np.pi)))

            component = gauss.Gaussian(prec=prec_i, h_vec=h_i, g_val=g_i, var_names=gaussian_mixture_a.components[0].var_names)
            resulting_gaussian_components.append(component)
        return GaussianMixture(resulting_gaussian_components)

    @staticmethod
    def _gm_division_m2(gma, gmb):
        """
        Method 2 for approximating the quotient (gma/gmb) of two Gaussian mixtures as a Gaussian mixture.
        This function does applies the moment matching equations to sets of negative definite mixtures (using
        _moment_match_complex_gaussian). For each of these mixtures, a single negative definite 'Gaussian' (h(x)) is then
        obtained. This function is then inverted (g(x) = 1/h(x)), resulting in a positive definite Gaussian function.
        The Gaussian mixture quotient (gma/gmb) is then approximated as the sum of the g(x) functions.

        This method is can be much faster than _gm_division_m1 (~10 times [tested on for one dimensional functions]),
        but also results a Gaussian mixture with the same number of components as gma. Whereas _gm_division_m2 should
        have the same number of components as the modes of the quotient function.

        :param GaussianMixture gma: The numerator factor
        :param GaussianMixture gmb: The denominator factor
        :returns: An approximation of the quotient function as a Gaussian mixture
        :rtype: GaussianMixture
        """
        # TODO: add check for dimensions that are not devided by
        #       - the variances in these will not change (if this makes sense)

        inverse_gaussian_mixtures = []
        resulting_gaussian_components = []
        count = 0
        for gaussian_a in gma.components:
            inverse_components = []
            for gaussian_b in gmb.components:
                conditional = gaussian_b.divide(gaussian_a)
                inverse_components.append(conditional)
            inverse_gaussian_mixture = GaussianMixture(inverse_components)
            inverse_gaussian_mixtures.append(inverse_gaussian_mixture)
            inverse_gaussian_approx = inverse_gaussian_mixture._moment_match_complex_gaussian()
            if _factor_utils.is_pos_def(inverse_gaussian_approx.prec):
                count += 1
                # raise ValueError(' precision is negative definite.')
                print("Warning: precision is negative definite.")
            gaussian_approx = inverse_gaussian_approx._invert()
            resulting_gaussian_components.append(gaussian_approx)
        print("num pos def = ", count, "/", len(gma.components))
        resulting_gm = GaussianMixture(resulting_gaussian_components)
        return resulting_gm

    def _get_limits_for_2d_plot(self):
        """
        Get x and y limits which includes most of the Gaussian mixture's mass, by considering
        the mean and variance of each Gaussian component.
        """
        x_lower_candidates = []
        x_upper_candidates = []
        y_lower_candidates = []
        y_upper_candidates = []
        for comp in self.components:
            stddev_x = np.sqrt(comp.get_cov()[0, 0])
            stddev_y = np.sqrt(comp.get_cov()[1, 1])
            mean_x = comp.get_mean()[0, 0]
            mean_y = comp.get_mean()[1, 0]
            x_lower_candidates.append(mean_x - 3.0 * stddev_x)
            x_upper_candidates.append(mean_x + 3.0 * stddev_x)
            y_lower_candidates.append(mean_y - 3.0 * stddev_y)
            y_upper_candidates.append(mean_y + 3.0 * stddev_y)
        x_lower = min(x_lower_candidates)
        x_upper = max(x_upper_candidates)
        y_lower = min(y_lower_candidates)
        y_upper = max(y_upper_candidates)
        return [x_lower, x_upper], [y_lower, y_upper]

    def _plot_2d(self, log, xlim, ylim):
        """
        Plot a 2d Gaussian mixture potential function

        :param log: if this is True, the log-potential will be plotted
        :param xlim: the x limits to plot the function over (optional)
        :param ylim: the y limits to plot the function over (optional)
        """
        if xlim is None and ylim is None:
            xlim, ylim = self._get_limits_for_2d_plot()
        elif xlim is not None and ylim is not None:
            pass
        else:
            print("Warning: partial limits received. Generating limits automatically.")
            xlim, ylim = self._get_limits_for_2d_plot()

        xlabel = self.var_names[0]
        ylabel = self.var_names[1]

        if not log:
            _factor_utils.plot_2d(func=self.potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)
        else:
            _factor_utils.plot_2d(func=self.log_potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)



"""
This module contains an abstract parent class, defining the some common default functions for factor template classes.
"""


from abc import ABC


class FactorTemplate(ABC):
    """
    A base class for factor templates.
    """

    def __init__(self, var_templates=None, conditioning_var_templates=None, conditional_var_templates=None):
        """
        The initializer.

        :param var_templates: The variable string templates (i.e: ['a_{i}', 'b{i}'])
        :type var_templates: str list
        :param conditioning_var_templates: The conditioning variable string templates (i.e: ['a_{i}', 'b{i}'])
        :param conditional_var_templates:  The conditional variable string templates (i.e: ['a_{i}', 'b{i}'])
        """
        self._var_templates = None
        self._conditioning_var_templates = None
        self._conditional_var_templates = None
        if (conditioning_var_templates is None) and (conditional_var_templates is None):
            self._var_templates = var_templates
        elif (conditioning_var_templates is None) or (conditional_var_templates is None):
            error_msg = (
                "neither or both of conditioning_var_templates and conditional_var_templates should be None"
            )
            raise ValueError(error_msg)
        else:
            self._conditioning_var_templates = conditioning_var_templates
            self._conditional_var_templates = conditional_var_templates

    def formattable(self, format_dict):
        """
        Check if the different variable templates sets are formattable by the given format_dict.

        :param format_dict: The dictionary used to format the variable templates.
        :return: The result of the check.
        :rtype: bool
        """
        # pylint: disable=expression-not-assigned
        try:
            if self._var_templates is not None:
                [vt.format(**format_dict) for vt in self._var_templates]
            else:
                [vt.format(**format_dict) for vt in self._conditioning_var_templates]
                [vt.format(**format_dict) for vt in self._conditional_var_templates]
            return True
        # pylint: enable=expression-not-assigned
        except KeyError:
            return False



"""
A module containing general functions to help with factor operations and transformations.
"""
# Standard imports
import copy

# Third-party imports
import numpy as np
import matplotlib.pyplot as plt


# TODO: simplify make_square_matrix and make_column vector (maybe renam to make ____ from list - there is already ...
# TODO: ... such a function for msquare matrix). These functions do not need al tie current functionality. ...
# TODO: ... Ensure factor classes still work.


def format_list_elements(str_list, format_dict):
    """
    Format a list of formattable stings.

    :param list str_list: The list of strings.
    :param format_dict: The dictionary specifying the values of the formattable parts of the strings.
    :return: A list of formatted
    :rtype: str list

    :Example:

        >>> format_list_elements(['a_{i}', 'b_{i}', 'c_{i}'], {'i':0})
        ['a_0', 'b_0', 'c_0']

    """
    formatted_list = [element.format(**format_dict) for element in str_list]
    return formatted_list


def get_subset_evidence(all_evidence_dict, subset_vars):
    """
    Select evidence for certain variables only from a evidence_dict and return the variables for which there is evidence
     and the corresponding values.

    :param all_evidence_dict: (dict) The evidence dictionary.
    :param subset_vars: (string list) the subset of variables for which to select evidence from all_evidence_dict.
    :return: evidence variable names and values
    """
    factor_scope_evidence = {v: all_evidence_dict[v] for v in subset_vars if v in all_evidence_dict}
    evidence_vrs = list(factor_scope_evidence.keys())
    evidence_values = list(factor_scope_evidence.values())
    return evidence_vrs, evidence_values


def is_pos_def(matrix):
    """
    Check if matrix is positive definite.

    :param matrix: the matrix to check
    :return: (bool) whether the matrix is positive-definite or not
    """
    return np.all(np.linalg.eigvals(matrix) > 0)


def remove_duplicate_values(array_like, tol=0.0):
    """
    Removes duplicate values from list (when tol=0.0) or remove approximately duplicate
    values if tol!=0.0.
    """
    unique_values = [array_like[0]]
    for element in array_like:
        element_is_duplicate = False
        for uval in unique_values:
            if abs(uval - element) <= tol:
                element_is_duplicate = True
        if not element_is_duplicate:
            unique_values.append(element)
    return unique_values


def make_scalar(scalar_like_object):
    """
    Convert a list or numpy array containing a single element to a scalar.

    :param scalar_like_object: a list or numpy array containing a single element (either int or float).
    :return: the int or float contained in the scalar_like_object
    """
    if isinstance(scalar_like_object, np.ndarray):
        if len(scalar_like_object) != 1:
            raise ValueError("array must have length of one to be convertible to a scalar.")
        return scalar_like_object.min()
    if isinstance(scalar_like_object, list):
        if len(scalar_like_object) != 1:
            raise ValueError("list must have length of one to be convertible to a scalar.")
        return scalar_like_object[0]
    if isinstance(scalar_like_object, (int, float)):
        return scalar_like_object
    raise TypeError(f"{type(scalar_like_object)} is not supported")


def make_column_vector(vector_like_object):
    """
    This function converts a vector like object into a standard
    numpy column vector.

    :param vector_like_object: an int, float, list or (one dimensional) array
    :return: a numpy column vector (array)
    """
    array = copy.deepcopy(vector_like_object)
    if isinstance(array, np.ndarray):
        array = array.ravel()
        error_msg = f"Error: array of shape {vector_like_object.shape} cannot be convertes to column vector"
        if len(vector_like_object.shape) > 1:
            assert not all([s > 1 for s in vector_like_object.shape]), error_msg
    else:
        array = np.array(vector_like_object).ravel()
    return np.expand_dims(array, axis=1)


# TODO: simplify this function like the one above
def list_to_square_matrix(matrix_like_object):
    """
    Make a square matrix from a list.

    :param matrix_like_object: a list containing a single int or float or n lists of length n
    :return: an nxn numpy array.
    """
    len_list = len(matrix_like_object)
    if len_list == 0:
        raise ValueError("cannot make matrix from empty list.")
    if len_list > 1:
        for element in matrix_like_object:
            if not isinstance(element, list):
                raise ValueError("cannot make square matrix from one dimensional list.")
            if len(element) != len_list:
                raise ValueError("cannot make square matrix from different length lists.")
            # if isinstance(element[0], list):
            #    raise ValueError('cannot make square matrix from 3d list')
        matrix = np.array(matrix_like_object)
    # len_list == 1:
    elif isinstance(matrix_like_object[0], (int, float)):
        # matrix_like_object = [float or int]
        matrix = np.array([matrix_like_object])
    elif len(matrix_like_object[0]) == 1:
        if isinstance(matrix_like_object[0][0], (int, float)):
            # matrix_like_object = [[float or int]]
            matrix = np.array(matrix_like_object)
    else:
        raise ValueError(f"cannot convert matrix_like_object {matrix_like_object} to square matrix")
    if (matrix.shape[0] != matrix.shape[1]) or len(matrix.shape) > 2:
        raise ValueError(f"cannot convert {matrix_like_object} to a square matrix")
    return matrix


def make_square_matrix(matrix_like_object):
    """
    This function converts a matrix like object into a standard numpy square matrix.

    :param matrix_like_object: an int, float, list of lists or (square, two dimensional) array
    :return: a numpy square matrix (array)
    """
    if isinstance(matrix_like_object, (int, float)):
        return np.array([[matrix_like_object]])
    if isinstance(matrix_like_object, list):
        return list_to_square_matrix(matrix_like_object)

    if isinstance(matrix_like_object, np.ndarray):
        if len(matrix_like_object.shape) != 2:
            raise ValueError("cannot convert non 2-dimensional array to matrix")
        assert matrix_like_object.shape[0] == matrix_like_object.shape[1]
        return matrix_like_object.copy()
    raise ValueError(f"cannot convert matrix_like_object {matrix_like_object} to square matrix.")


def indexed_square_matrix_operation(mat_a, mat_b, var_names_a, var_names_b, operator):
    """
    A function that performs element wise operations between square matrices where the corresponding indices in the
    different matrices do not necessarily correspond to the same variables, but can be mapped to each other through the
    variable name lists (var_names_a and var_names_b).

    :param mat_a: A nxn numpy array to be used with the operator
    :param mat_b:  A nxn numpy array to be used with the operator
    :param var_names_a: the variable names corresponding to the values in mat_a
    :param var_names_b: the variable names corresponding to the values in mat_b
    :param operator: the operator to be applied to mat_a and mat_b
    :return: the result of the operation

    Example:
        With:
            mat_a =
            [[11, 12, 13],
             [21, 22, 23],
             [31, 32, 33]]
            var_names_a = ['var1', 'var2']
            mat_b =
            [[11 , 13],
             [31,  33]]
            var_names_b = ['var1', 'var3']
            operator_str = '-'
        the function returns:
        [[0,  12, 0],
         [21, 22, 23],
         [0,  32, 0]]
    """

    assert len(mat_a.shape) == 2, "Error: matrix needs to be 2 dimensional"
    assert len(mat_b.shape) == 2, "Error: matrix needs to be 2 dimensional"
    assert mat_a.shape[0] == mat_a.shape[1], "Error: matrix needs to be square"
    assert mat_b.shape[0] == mat_b.shape[1], "Error: matrix needs to be square"
    var_names_len_error_msg = (
        "The number of variables in var_names_{aorb} does not match the dimensions of mat_{aorb}."
    )
    assert len(var_names_a) == mat_a.shape[0], var_names_len_error_msg.format(aorb="a")
    assert len(var_names_b) == mat_b.shape[0], var_names_len_error_msg.format(aorb="b")

    # TODO: clean this up (also see todo in indexed_column_vector_operation)
    if set(var_names_b) <= set(var_names_a):
        new_vars = var_names_a
    elif set(var_names_a) <= set(var_names_b):
        new_vars = var_names_b
    else:
        new_vars = list(set(var_names_b + var_names_a))
        new_vars.sort()
    new_dim = len(new_vars)
    new_matrix = np.zeros([new_dim, new_dim])

    for i, var_i_name in enumerate(new_vars):
        for j, var_j_name in enumerate(new_vars):
            val_a = 0.0
            if (var_i_name in var_names_a) and (var_j_name in var_names_a):
                val_a = mat_a[var_names_a.index(var_i_name), var_names_a.index(var_j_name)]
            val_b = 0.0
            if (var_i_name in var_names_b) and (var_j_name in var_names_b):
                val_b = mat_b[var_names_b.index(var_i_name), var_names_b.index(var_j_name)]
            new_matrix[i, j] = operator(val_a, val_b)

    return new_matrix, new_vars


def indexed_column_vector_operation(colvec_a, colvec_b, var_names_a, var_names_b, operator):
    """
    A function that performs element operations between column vectors, where the corresponding indices in the different
    vectors do not necessarily correspond to the same variables, but can be mapped to each other through the variable
    name lists (var_names_a and var_names_b).

    :param colvec_a: A nx1 numpy array to be used with the operator
    :param colvec_b:  A nx1 numpy array to be used with the operator
    :param var_names_a: the variable names corresponding to the values in colvec_a
    :param var_names_b: the variable names corresponding to the values in colvec_b
    :param operator: the operator to be applied to colvec_a and colvec_b
    :return: the result of the operation

    Example:
        With:
            colvec_a =
            [[11],
             [21],
             [31]]
            var_names_a = ['var1', 'var2']
            colvec_b =
            [[11],
             [31]]
            var_names_b = [var1', 'var3']
            operator_str = '-'
        the function returns:
        [[0],
         [21],
         [0]]
    """

    assert len(colvec_a.shape) == 2, "Error: column vector needs to be 2 dimensional"
    assert len(colvec_b.shape) == 2, "Error: column vector needs to be 2 dimensional"
    assert colvec_a.shape[1] == 1, "Error: column vector array cannot have more than one column"
    assert colvec_b.shape[1] == 1, "Error: column vector array cannot have more than one column"
    assert (
            len(var_names_a) == colvec_a.shape[0]
    ), "Error: number of variables in var_names_a does not match the dimension of colvec_a."
    assert (
            len(var_names_b) == colvec_b.shape[0]
    ), "Error: number of variables in var_names_a does not match the dimension of colvec_b."
    # TODO: clean this up (also see todo in indexed_square_matrix_operation)
    if set(var_names_b) <= set(var_names_a):
        new_vars = var_names_a
    elif set(var_names_a) <= set(var_names_b):
        new_vars = var_names_b
    else:
        new_vars = list(set(var_names_b + var_names_a))
        new_vars.sort()
    new_dim = len(new_vars)
    new_vector = np.zeros([new_dim, 1])

    for i, var_i_name in enumerate(new_vars):
        val_a = 0.0
        if var_i_name in var_names_a:
            val_a = colvec_a[var_names_a.index(var_i_name), 0]
        val_b = 0.0
        if var_i_name in var_names_b:
            val_b = colvec_b[var_names_b.index(var_i_name), 0]
        new_vector[i, 0] = operator(val_a, val_b)

    return new_vector, new_vars


def plot_2d(func, xlim, ylim, xlabel, ylabel, figsize=None):  # pragma: no cover
    """
    Plot a 2d function, with specific x limits and y limits.

    :param xlim: the x limits to plot the function over.
    :param ylim: the y limits to plot the function over.
    :param func: the function to plot
    :param xlim: the x limits over which to plot the function
    :param ylim: the x limits over which to plot the function
    :param xlabel: the x label for the plot
    :param ylabel: the y label for the plot
    :param figsize: the figure size
    """

    x_space = np.linspace(xlim[0], xlim[1], num=100)
    y_space = np.linspace(ylim[0], ylim[1], num=100)

    x_mesh, y_mesh = np.meshgrid(x_space, y_space)
    xx_mesh_array = np.array([x_mesh.ravel(), y_mesh.ravel()]).T
    # pylint: disable=not-an-iterable
    z_mesh_array = np.array([func(X) for X in xx_mesh_array])
    # pylint: enable=not-an-iterable
    z_mesh = z_mesh_array.reshape(x_mesh.shape)

    if figsize is not None:
        plt.figure(figsize=figsize)
    color_scale = plt.contourf(x_mesh, y_mesh, z_mesh, levels=30)
    plt.colorbar(color_scale, shrink=0.8, extend="both")

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)


def inv_matrix(mat):
    """
    Invert matrix using the numpy.linalg.inv function and describe matrix if inversion fails.

    :param numpy.ndarray mat: The matrix to invert.
    :return: The inverted matrix.
    :rtype: numpy.ndarray
    """
    try:
        mat_inv = np.linalg.inv(mat)
    except np.linalg.LinAlgError as error:
        raise type(error)(error.args[0] + f":\n{mat}")
    return mat_inv


def log(x_val):
    """ "
    Get the log of a value using the numpy.log function and output the input value if invalid.

    :param numerical x_val: The value to compute the log of.
    :return: The log value.
    :rtype: numerical
    """
    if x_val < 0.0:
        raise Warning(f"Invalid value ({x_val}) in log")
    logx = np.log(x_val)
    return logx


def tabs_to_cover_string(string):
    """
     Get the number of tabs required to be at least the same length as a given string.
     :param string: The string
     :return: The number of tabs to cover it
     :rtype: int
     """
    num_tabs = int(np.floor(len(string) / 8) + 1)
    return num_tabs


def space_assignments_and_probs(assignment, value, spacings):
    """
    Join the string representations of the assignment values and prob value, with the provided spacing.

    :param list assignment: The list of assignments typically corresponding to the entry in a categorical table.
    :param float value: The value corresponding to the assignment.
    :param spacings: The spacing to apply between the values.
    :return: The spaced, joined string
    :rtype: str
    """
    content_stings = [str(a) for a in assignment] + [f'{value:.4f}']
    spaced_content_stings = [c_str + space for c_str, space in zip(content_stings, spacings)]
    spaced_content_sting = ''.join(spaced_content_stings)
    return spaced_content_sting



import numpy as np
from ipywidgets import interact
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Ellipse

"""
A helper module for the Kalman filter example notebook.
"""


def _get_ellipse_patch(cov, mean, std_devs=2):
    eigen_values_array, eigen_vectors_array = np.linalg.eig(cov)
    max_len_eig_vector = eigen_vectors_array[:, np.argmax(eigen_values_array)]
    angle = np.arctan2(max_len_eig_vector[1], max_len_eig_vector[0])
    std_dev_primary_comp = np.sqrt(eigen_values_array[0])
    std_dev_secondary_comp = np.sqrt(eigen_values_array[1])

    ellipse = matplotlib.patches.Ellipse(mean,
                                         std_dev_primary_comp * std_devs,
                                         std_dev_secondary_comp * std_devs,
                                         angle=angle,
                                         fc=(1, 0, 0, 0),
                                         ec=(0, 0, 1, 1),
                                         lw=2)
    return ellipse


def _add_ellipse(ax, ellipse):
    ax.add_patch(ellipse)


def _get_2d_std_dev_ellipses(gaussians):
    ellipse_patches = []
    for g in gaussians:
        ellipse_patch = _get_ellipse_patch(g.get_cov(), g.get_mean())
        ellipse_patches.append(ellipse_patch)
    return ellipse_patches


# TODO: Consider removing this.
def _plot_2d_std_dev_ellipses(gaussians):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    patches = _get_2d_std_dev_ellipses(gaussians)
    for patch in patches:
        ax.add_patch(patch)
    ax.autoscale()


# TODO: The widget breaks when moving the slider in reverse. Fix this.
def inferred_system_state_widget(X_dist_over_time, X_meas_over_time, X_true_over_time):
    """
    Plot a series of two dimensional Gaussian state distributions (as standard-deviation ellipses) together with the
    ground truth state and state-space measurements.
    :param X_dist_over_time: The time evolution series of the state distributions.
    :type X_dist_over_time: Gaussian list
    :param X_meas_over_time: The measurements over time.
    :type X_meas_over_time: array like list
    :param X_true_over_time: The ground truth time evolution series of the state.
    :type X_true_over_time: array like list
    """
    all_xs = X_meas_over_time[:, 0].tolist() + X_true_over_time[:, 0].tolist()
    all_ys = X_meas_over_time[:, 1].tolist() + X_true_over_time[:, 1].tolist()
    ylims = [min(all_ys), max(all_ys)]
    xlims = [min(all_xs), max(all_xs)]

    height = ylims[1] - ylims[0]
    width = xlims[1] - xlims[0]

    legend_elements = [Line2D([0], [0], marker='o', color='w', label='Scatter', markerfacecolor='b', markersize=15),
                       Line2D([0], [0], marker='o', color='w', label='Scatter', markerfacecolor='r', markersize=15),
                       Line2D([0], [0], marker='o', color='w', label='Scatter', markerfacecolor='k', markersize=15)]

    estimated_cov_ellipses = _get_2d_std_dev_ellipses(X_dist_over_time)
    estimated_means_array = np.concatenate([p.get_mean().T for p in X_dist_over_time])
    last_timestep_index = len(estimated_cov_ellipses) - 1

    def update(t=0):
        """Remove old lines from plot and plot new one"""
        fig_width = 20
        fig, ax = plt.subplots(figsize=(fig_width, fig_width * height / width))
        ax.set_ylim(ylims)
        ax.set_xlim(xlims)
        ax.grid(True)
        [l.remove() for l in ax.patches]
        ax.add_patch(estimated_cov_ellipses[t])
        plt.scatter(estimated_means_array[t:t + 1, 0], estimated_means_array[t:t + 1, 1], color='b')
        plt.scatter(X_meas_over_time[t - 1, 0], X_meas_over_time[t - 1, 1], color='r')
        plt.plot(X_true_over_time[:t, 0], X_true_over_time[:t, 1], color='k', marker='o')
        plt.legend(legend_elements, ['estimate', 'measurement', 'truth'])
        plt.show()

    interact(update, t=(1, last_timestep_index-1));
